import unittest
from typing import Optional, Dict, Set, List, Any
from uuid import uuid4

from StructNoSQL import BaseField, MapModel, TableDataModel
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    class ContainerModel(MapModel):
        untypedList = BaseField(field_type=list, key_name='listIndex', required=False)
        typedList = BaseField(field_type=List[str], key_name='listIndex', required=False)
    container = BaseField(field_type=ContainerModel, required=False)

class TestListObjectType(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_untyped_list(self):
        random_list_values: List[Any] = list(str(uuid4()) for i in range(5))
        single_valid_list_item = random_list_values[2]
        random_list_values.append(42)
        # Add a value of different type, that is valid to the untypedList

        set_success: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='container.untypedList',
            value_to_set=random_list_values
        )
        self.assertTrue(set_success)

        retrieved_list_item: Optional[str or int] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='container.untypedList.{{listIndex}}',
            query_kwargs={'listIndex': 2}
        )
        self.assertEqual(retrieved_list_item, single_valid_list_item)

        retrieved_entire_list: Optional[list] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='container.untypedList',
            query_kwargs={'listIndex': 2}
        )
        self.assertEqual(retrieved_entire_list, random_list_values)

        deletion_success: bool = self.users_table.delete_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='container.untypedList'
        )
        self.assertTrue(deletion_success)

    def test_typed_list(self):
        valid_random_list_values: List[str] = [str(uuid4()) for i in range(5)]
        single_valid_list_item = valid_random_list_values[2]
        random_list_values = valid_random_list_values.copy()
        random_list_values.append(42)
        # Add invalid value to the random_list_values

        set_success: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='container.typedList',
            value_to_set=random_list_values
        )
        self.assertTrue(set_success)

        query_kwargs = {'listIndex': 2}
        retrieved_list_item: Optional[dict] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='container.typedList.{{listIndex}}',
            query_kwargs=query_kwargs
        )
        self.assertEqual(retrieved_list_item, single_valid_list_item)

        retrieved_entire_list: Optional[list] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='container.typedList',
            query_kwargs=query_kwargs
        )
        self.assertEqual(valid_random_list_values, retrieved_entire_list)

        deletion_success: bool = self.users_table.delete_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='container.typedList',
            query_kwargs=query_kwargs
        )
        self.assertTrue(deletion_success)


if __name__ == '__main__':
    unittest.main()



from typing import Any, Optional

from StructNoSQL.servers_middlewares.external_dynamodb_api.api_executor_handlers import getSingleValueInPathTarget, \
    getValuesInMultiplePathTarget, setUpdateMultipleDataElementsToMap, setUpdateMultipleDataElementsToMapWithDefaultInitialization,\
    removeDataElementsFromMap, deleteDataElementsFromMap, getOrQuerySingleItem, deleteRecord, removeRecord, putRecord, queryItemsByKey


class StructNoSQLParser:
    HANDLERS_SWITCH = {
        'putRecord': putRecord,
        'deleteRecord': deleteRecord,
        'removeRecord': removeRecord,
        'getOrQuerySingleItem': getOrQuerySingleItem,
        'getSingleValueInPathTarget': getSingleValueInPathTarget,
        'getValuesInMultiplePathTarget': getValuesInMultiplePathTarget,
        'setUpdateMultipleDataElementsToMap': setUpdateMultipleDataElementsToMap,
        'setUpdateMultipleDataElementsToMapWithDefaultInitialization': setUpdateMultipleDataElementsToMapWithDefaultInitialization,
        'removeDataElementsFromMap': removeDataElementsFromMap,
        'deleteDataElementsFromMap': deleteDataElementsFromMap,
        'queryItemsByKey': queryItemsByKey
    }

    @staticmethod
    def parse(data: dict) -> Optional[Any]:
        operation_type: Optional[str] = data.get('operationType', None)
        if operation_type is None or not isinstance(operation_type, str):
            return None

        handler = StructNoSQLParser.HANDLERS_SWITCH.get(operation_type, None)
        if handler is None:
            print(f"No handler found for {operation_type}")
            return None

        return handler



import re
from typing import List, Dict, Any, Optional, Tuple, Union
from StructNoSQL.models import DatabasePathElement
from StructNoSQL.exceptions import FieldTargetNotFoundException
from StructNoSQL.fields import BaseItem, BaseField, MapModel
from StructNoSQL.practical_logger import message_with_vars
from StructNoSQL.exceptions import MissingQueryKwarg

MULTI_ATTRIBUTES_SELECTOR_REGEX_EXPRESSION = r'(\()(.*)(\))'


def _get_field_object_from_field_path(field_path_key: str, fields_switch: Dict[str, BaseField]) -> BaseField:
    current_field_object: Optional[BaseField] = fields_switch.get(field_path_key, None)
    if current_field_object is not None:
        return current_field_object
    else:
        raise FieldTargetNotFoundException(message_with_vars(
            message=f"A field target to get was not found.",
            vars_dict={"field_path_key": field_path_key, "fields_switch": fields_switch}
        ))

def process_and_get_field_path_object_from_field_path(field_path_key: str, fields_switch: dict) -> Tuple[Union[BaseField, Dict[str, BaseField]], bool]:
    matches: Optional[List[tuple]] = re.findall(pattern=MULTI_ATTRIBUTES_SELECTOR_REGEX_EXPRESSION, string=field_path_key)
    if matches is not None and len(matches) > 0:
        for match in matches:
            selected_string = ''.join([selector for selector in match])
            attributes_selectors_string: str = match[1]
            attributes_selector_list = attributes_selectors_string.replace(' ', '').split(',')

            attributes_fields_objets: Dict[str, BaseField] = {}
            for attribute_selector in attributes_selector_list:
                current_attribute_field_path = field_path_key.replace(selected_string, attribute_selector)
                """if len(current_attribute_field_path) > 0 and current_attribute_field_path[0] == ".":
                    # If the start of the field path is a multi selector, the replace will unfortunately add an
                    # invalid point, which we will remove if we see that the first char of the field path is a point.
                    current_attribute_field_path = current_attribute_field_path[1:]"""

                attributes_fields_objets[attribute_selector] = _get_field_object_from_field_path(
                    field_path_key=current_attribute_field_path, fields_switch=fields_switch
                )

            num_attributes_fields: int = len(attributes_fields_objets)
            if not num_attributes_fields > 0:
                raise Exception(message_with_vars(
                    message="Cannot use an attribute selector ( ) without specifying any attribute inside it.",
                    vars_dict={'field_path_key': field_path_key, 'attributes_selectors_string': attributes_selectors_string}
                ))
            return attributes_fields_objets, True

    return _get_field_object_from_field_path(field_path_key=field_path_key, fields_switch=fields_switch), False


def make_rendered_database_path(database_path_elements: List[DatabasePathElement], query_kwargs: dict) -> List[DatabasePathElement]:
    output_database_path_elements: List[DatabasePathElement] = []
    for path_element in database_path_elements:
        if "$key$:" not in path_element.element_key:
            # If the path_element do not contains a key that need to be modified, we can use the current
            # instance of the path element, since it will not be modified, and so will not cause issue
            # when other invocations of queries and operations will use the same path element instance.
            output_database_path_elements.append(path_element)
        else:
            variable_name = path_element.element_key.replace('$key$:', '')
            if query_kwargs is not None:
                matching_kwarg: Optional[Any] = query_kwargs.get(variable_name, None)
                if matching_kwarg is not None:
                    # If the key of the path_element needs to be modified, we do not modify the existing path element,
                    # but we create a new instance of path element. Since the database_path_elements variable is retrieved
                    # using the static _database_path variable, the path elements in database_path_elements needs to
                    # remained unmodified, so that other invocations of queries and operations will be able to have
                    # cleans element keys that will properly be filled with the query_kwargs specified in the request.
                    output_database_path_elements.append(DatabasePathElement(
                        element_key=matching_kwarg,
                        default_type=path_element.default_type,
                        custom_default_value=path_element.custom_default_value
                    ))
                else:
                    raise MissingQueryKwarg(message_with_vars(
                        message="A variable was required but not found in the query_kwargs dict passed to the make_rendered_database_path function.",
                        vars_dict={"keyVariableName": variable_name, "matchingKwarg": matching_kwarg,
                                   "queryKwargs": query_kwargs, "databasePathElements": database_path_elements}
                    ))
            else:
                raise Exception(message_with_vars(
                    message="A variable was required but no query_kwargs have been passed to the make_rendered_database_path function.",
                    vars_dict={"keyVariableName": variable_name, "queryKwargs": query_kwargs, "databasePathElements": database_path_elements}
                ))
    return output_database_path_elements

def process_and_make_single_rendered_database_path(field_path: str, fields_switch: dict, query_kwargs: dict) -> Tuple[
    Union[
        Tuple[BaseField, List[DatabasePathElement]],
        Dict[str, Tuple[BaseField, List[DatabasePathElement]]]
    ],
    bool
]:
    field_path_object, is_multi_selector = process_and_get_field_path_object_from_field_path(
        field_path_key=field_path, fields_switch=fields_switch
    )
    if is_multi_selector is not True:
        field_path_object: BaseField
        rendered_database_path_elements: List[DatabasePathElement] = make_rendered_database_path(
            database_path_elements=field_path_object.database_path, query_kwargs=query_kwargs
        )
        return (field_path_object, rendered_database_path_elements), False
    else:
        field_path_object: Dict[str, BaseField]
        fields_rendered_database_path_elements: Dict[str, Tuple[BaseField, List[DatabasePathElement]]] = {}
        for single_field_path_object in field_path_object.values():
            rendered_database_path_elements: List[DatabasePathElement] = make_rendered_database_path(
                database_path_elements=single_field_path_object.database_path, query_kwargs=query_kwargs
            )
            fields_rendered_database_path_elements[single_field_path_object.field_name] = (single_field_path_object, rendered_database_path_elements)
        return fields_rendered_database_path_elements, True

def process_validate_data_and_make_single_rendered_database_path(
        field_path: str, fields_switch: dict, query_kwargs: dict, data_to_validate: Any
) -> Tuple[BaseField, List[DatabasePathElement], Optional[Any], bool]:

    # todo: add support for multiple fields path
    field_object, is_multi_selector = process_and_get_field_path_object_from_field_path(
        field_path_key=field_path, fields_switch=fields_switch
    )
    rendered_database_path_elements: List[DatabasePathElement] = make_rendered_database_path(
        database_path_elements=field_object.database_path, query_kwargs=query_kwargs
    )

    field_object.populate(value=data_to_validate)
    validated_data, valid = field_object.validate_data()
    if valid is True:
        return field_object, rendered_database_path_elements, validated_data, True
    else:
        return field_object, rendered_database_path_elements, None, False

def join_field_path_elements(field_path_elements: List[DatabasePathElement]) -> str:
    return '.'.join((f'{item.element_key}' for item in field_path_elements))



import unittest
from typing import Optional, Union, Dict, Any
from uuid import uuid4

from StructNoSQL import DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiCachingTable, FieldSetter
from StructNoSQL.clients_middlewares.external_dynamodb_api.external_dynamodb_api_basic_table import ExternalDynamoDBApiBasicTable
from tests.components.playground_table_clients import TEST_ACCOUNT_ID


def test_update_field_return_old(
        self: unittest.TestCase, users_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        primary_key_name: str, is_caching: bool
):
    simple_text_field_first_random_value: str = f"simpleTextField_firstRandomValue_{uuid4()}"
    first_random_value_update_success: bool = users_table.update_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleTextField', value_to_set=simple_text_field_first_random_value
    )
    self.assertTrue(first_random_value_update_success)

    if is_caching is True:
        operations_commit_success: bool = users_table.commit_operations()
        self.assertTrue(operations_commit_success)
        users_table.clear_cached_data()

    simple_text_field_second_random_value: str = f"simpleTextField_secondRandomValue_{uuid4()}"
    second_random_value_update_success, field_old_value = users_table.update_field_return_old(
        key_value=TEST_ACCOUNT_ID, field_path='simpleTextField', value_to_set=simple_text_field_second_random_value
    )
    self.assertTrue(second_random_value_update_success)
    self.assertEqual((
        simple_text_field_first_random_value if is_caching is not True else
        {'fromCache': False, 'value': simple_text_field_first_random_value}
    ), field_old_value)

    if is_caching is True:
        operations_commit_success: bool = users_table.commit_operations()
        self.assertTrue(operations_commit_success)
        users_table.clear_cached_data()

    retrieved_second_random_value: Optional[str] = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='simpleTextField')
    self.assertEqual((
        simple_text_field_second_random_value if is_caching is not True else
        {'fromCache': False, 'value': simple_text_field_second_random_value}
    ), retrieved_second_random_value)


def test_update_multiple_fields_return_old(
        self: unittest.TestCase, users_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        primary_key_name: str, is_caching: bool
):
    container_field_one_first_random_value: str = f"container_fieldOne_firstRandomValue_{uuid4()}"
    container_field_two_first_random_value: str = f"container_fieldTwo_firstRandomValue_{uuid4()}"
    container_fields_first_update_success: bool = users_table.update_multiple_fields(
        key_value=TEST_ACCOUNT_ID, setters=[
            FieldSetter(field_path='container.textFieldOne', value_to_set=container_field_one_first_random_value),
            FieldSetter(field_path='container.textFieldTwo', value_to_set=container_field_two_first_random_value)
        ]
    )
    self.assertTrue(container_fields_first_update_success)

    if is_caching is True:
        operations_commit_success: bool = users_table.commit_operations()
        self.assertTrue(operations_commit_success)
        users_table.clear_cached_data()

    container_field_one_second_random_value: str = f"container_fieldOne_secondRandomValue_{uuid4()}"
    container_field_two_second_random_value: str = f"container_fieldTwo_secondRandomValue_{uuid4()}"
    container_fields_second_update_success, field_old_value = users_table.update_multiple_fields_return_old(
        key_value=TEST_ACCOUNT_ID, setters={
            'fieldOne': FieldSetter(field_path='container.textFieldOne', value_to_set=container_field_one_second_random_value),
            'fieldTwo': FieldSetter(field_path='container.textFieldTwo', value_to_set=container_field_two_second_random_value)
        }
    )
    self.assertTrue(container_fields_second_update_success)
    self.assertEqual({
        'fieldOne': (
            container_field_one_first_random_value if is_caching is not True else
            {'fromCache': False, 'value': container_field_one_first_random_value}
        ),
        'fieldTwo': (
            container_field_two_first_random_value if is_caching is not True else
            {'fromCache': False, 'value': container_field_two_first_random_value}
        )
    }, field_old_value)

    if is_caching is True:
        users_table.clear_cached_data()

    retrieved_container_fields_second_random_values: Optional[Dict[str, Optional[Any]]] = users_table.get_field(
        key_value=TEST_ACCOUNT_ID, field_path='container.(textFieldOne, textFieldTwo)'
    )
    self.assertEqual({
        'textFieldOne': (
            container_field_one_second_random_value if is_caching is not True else
            {'fromCache': False, 'value': container_field_one_second_random_value}
        ),
        'textFieldTwo': (
            container_field_two_second_random_value if is_caching is not True else
            {'fromCache': False, 'value': container_field_two_second_random_value}
        )
    }, retrieved_container_fields_second_random_values)



import unittest
from typing import Optional, Union, Dict, Any, Tuple
from uuid import uuid4

from StructNoSQL import DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiCachingTable, FieldSetter, \
    FieldGetter, FieldRemover
from StructNoSQL.clients_middlewares.external_dynamodb_api.external_dynamodb_api_basic_table import ExternalDynamoDBApiBasicTable
from tests.components.playground_table_clients import TEST_ACCOUNT_ID


def test_get_field(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    simple_field_random_text_value: str = f"simpleField_randomTextValue_{uuid4()}"
    first_table_simple_field_update_success: bool = first_table.update_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', value_to_set=simple_field_random_text_value
    )
    self.assertTrue(first_table_simple_field_update_success)
    if is_caching is True:
        self.assertTrue(first_table.commit_operations())
        first_table.clear_cached_data()

    second_table_retrieved_simple_field_without_data_validation: Optional[Any] = second_table.get_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', data_validation=False
    )
    self.assertEqual((
        simple_field_random_text_value if is_caching is not True else
        {'value': simple_field_random_text_value, 'fromCache': False}
    ), second_table_retrieved_simple_field_without_data_validation)

    if is_caching is True:
        second_table.clear_cached_data()

    second_table_retrieved_simple_field_with_data_validation: Optional[int] = second_table.get_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', data_validation=True
    )
    self.assertEqual((
        None if is_caching is not True else
        {'value': None, 'fromCache': False}
    ), second_table_retrieved_simple_field_with_data_validation)


def test_get_field_multi_selectors(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    container_field_one_random_text_value: str = f"container_fieldOne_randomTextValue_{uuid4()}"
    container_field_two_random_text_value: str = f"container_fieldTwo_randomTextValue_{uuid4()}"
    first_table_container_fields_update_success: bool = first_table.update_multiple_fields(
        key_value=TEST_ACCOUNT_ID, setters=[
            FieldSetter(field_path='container.nestedFieldOne', value_to_set=container_field_one_random_text_value),
            FieldSetter(field_path='container.nestedFieldTwo', value_to_set=container_field_two_random_text_value)
        ]
    )
    self.assertTrue(first_table_container_fields_update_success)
    if is_caching is True:
        self.assertTrue(first_table.commit_operations())
        first_table.clear_cached_data()

    second_table_retrieved_container_fields_without_data_validation: Dict[str, Optional[Any]] = second_table.get_field(
        key_value=TEST_ACCOUNT_ID, field_path='container.(nestedFieldOne, nestedFieldTwo)', data_validation=False
    )
    self.assertEqual({
        'nestedFieldOne': (
            container_field_one_random_text_value if is_caching is not True else
            {'value': container_field_one_random_text_value, 'fromCache': False}
        ),
        'nestedFieldTwo': (
            container_field_two_random_text_value if is_caching is not True else
            {'value': container_field_two_random_text_value, 'fromCache': False}
        )
    }, second_table_retrieved_container_fields_without_data_validation)

    if is_caching is True:
        second_table.clear_cached_data()

    second_table_retrieved_container_fields_with_data_validation: Dict[str, Optional[int]] = second_table.get_field(
        key_value=TEST_ACCOUNT_ID, field_path='container.(nestedFieldOne, nestedFieldTwo)'
    )
    self.assertEqual({
        'nestedFieldOne': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        ),
        'nestedFieldTwo': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        )
    }, second_table_retrieved_container_fields_with_data_validation)


def test_get_multiple_fields(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    container_field_one_random_text_value: str = f"container_fieldOne_randomTextValue_{uuid4()}"
    container_field_two_random_text_value: str = f"container_fieldTwo_randomTextValue_{uuid4()}"
    first_table_container_fields_update_success: bool = first_table.update_multiple_fields(
        key_value=TEST_ACCOUNT_ID, setters=[
            FieldSetter(field_path='container.nestedFieldOne', value_to_set=container_field_one_random_text_value),
            FieldSetter(field_path='container.nestedFieldTwo', value_to_set=container_field_two_random_text_value)
        ]
    )
    self.assertTrue(first_table_container_fields_update_success)

    if is_caching is True:
        self.assertTrue(first_table.commit_operations())
        first_table.clear_cached_data()

    second_table_retrieved_container_fields_without_data_validation: Dict[str, Optional[Any]] = second_table.get_multiple_fields(
        key_value=TEST_ACCOUNT_ID, data_validation=False, getters={
            'one': FieldGetter(field_path='container.nestedFieldOne'),
            'two': FieldGetter(field_path='container.nestedFieldTwo'),
        }
    )
    self.assertEqual({
        'one': (
            container_field_one_random_text_value if is_caching is not True else
            {'value': container_field_one_random_text_value, 'fromCache': False}
        ),
        'two': (
            container_field_two_random_text_value if is_caching is not True else
            {'value': container_field_two_random_text_value, 'fromCache': False}
        )
    }, second_table_retrieved_container_fields_without_data_validation)

    if is_caching is True:
        second_table.clear_cached_data()

    second_table_retrieved_container_fields_with_data_validation: Dict[str, Optional[int]] = second_table.get_multiple_fields(
        key_value=TEST_ACCOUNT_ID, data_validation=True, getters={
            'one': FieldGetter(field_path='container.nestedFieldOne'),
            'two': FieldGetter(field_path='container.nestedFieldTwo'),
        }
    )
    self.assertEqual({
        'one': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        ),
        'two': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        )
    }, second_table_retrieved_container_fields_with_data_validation)


def test_remove_field(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    def generate_update_simple_field_text_value() -> str:
        simple_field_random_text_value: str = f"simpleField_randomTextValue_{uuid4()}"
        first_table_simple_field_update_success: bool = first_table.update_field(
            key_value=TEST_ACCOUNT_ID, field_path='simpleField', value_to_set=simple_field_random_text_value
        )
        self.assertTrue(first_table_simple_field_update_success)
        if is_caching is True:
            self.assertTrue(first_table.commit_operations())
            first_table.clear_cached_data()
        return simple_field_random_text_value

    first_generated_simple_field_random_text_value: str = generate_update_simple_field_text_value()
    second_table_removed_simple_field_without_data_validation: Optional[Any] = second_table.remove_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', data_validation=False
    )
    self.assertEqual((
        first_generated_simple_field_random_text_value if is_caching is not True else
        {'value': first_generated_simple_field_random_text_value, 'fromCache': False}
    ), second_table_removed_simple_field_without_data_validation)

    if is_caching is True:
        second_table.clear_cached_data()

    second_generated_simple_field_random_text_value: str = generate_update_simple_field_text_value()
    second_table_removed_simple_field_with_data_validation: Optional[int] = second_table.remove_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', data_validation=True
    )
    self.assertEqual((
        None if is_caching is not True else
        {'value': None, 'fromCache': False}
    ), second_table_removed_simple_field_with_data_validation)


def test_remove_field_multi_selectors(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    def generate_update_container_fields_text_values() -> Tuple[str, str]:
        container_field_one_random_text_value: str = f"container_fieldOne_randomTextValue_{uuid4()}"
        container_field_two_random_text_value: str = f"container_fieldTwo_randomTextValue_{uuid4()}"
        first_table_container_fields_update_success: bool = first_table.update_multiple_fields(
            key_value=TEST_ACCOUNT_ID, setters=[
                FieldSetter(field_path='container.nestedFieldOne', value_to_set=container_field_one_random_text_value),
                FieldSetter(field_path='container.nestedFieldTwo', value_to_set=container_field_two_random_text_value)
            ]
        )
        self.assertTrue(first_table_container_fields_update_success)
        if is_caching is True:
            self.assertTrue(first_table.commit_operations())
            first_table.clear_cached_data()
        return container_field_one_random_text_value, container_field_two_random_text_value

    first_generated_container_field_one_random_text_value, first_generated_container_field_two_random_text_value = generate_update_container_fields_text_values()
    second_table_removed_container_fields_without_data_validation: Dict[str, Optional[Any]] = second_table.remove_field(
        key_value=TEST_ACCOUNT_ID, field_path='container.(nestedFieldOne, nestedFieldTwo)', data_validation=False
    )
    self.assertEqual({
        'nestedFieldOne': (
            first_generated_container_field_one_random_text_value if is_caching is not True else
            {'value': first_generated_container_field_one_random_text_value, 'fromCache': False}
        ),
        'nestedFieldTwo': (
            first_generated_container_field_two_random_text_value if is_caching is not True else
            {'value': first_generated_container_field_two_random_text_value, 'fromCache': False}
        )
    }, second_table_removed_container_fields_without_data_validation)

    if is_caching is True:
        second_table.clear_cached_data()

    second_generated_container_field_one_random_text_value, second_generated_container_field_two_random_text_value = generate_update_container_fields_text_values()
    second_table_removed_container_fields_with_data_validation: Dict[str, Optional[int]] = second_table.remove_field(
        key_value=TEST_ACCOUNT_ID, field_path='container.(nestedFieldOne, nestedFieldTwo)', data_validation=True
    )
    self.assertEqual({
        'nestedFieldOne': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        ),
        'nestedFieldTwo': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        )
    }, second_table_removed_container_fields_with_data_validation)


def test_remove_multiple_fields(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    def generate_update_container_fields_text_values() -> Tuple[str, str]:
        container_field_one_random_text_value: str = f"container_fieldOne_randomTextValue_{uuid4()}"
        container_field_two_random_text_value: str = f"container_fieldTwo_randomTextValue_{uuid4()}"
        first_table_container_fields_update_success: bool = first_table.update_multiple_fields(
            key_value=TEST_ACCOUNT_ID, setters=[
                FieldSetter(field_path='container.nestedFieldOne', value_to_set=container_field_one_random_text_value),
                FieldSetter(field_path='container.nestedFieldTwo', value_to_set=container_field_two_random_text_value)
            ]
        )
        self.assertTrue(first_table_container_fields_update_success)
        if is_caching is True:
            self.assertTrue(first_table.commit_operations())
            first_table.clear_cached_data()
        return container_field_one_random_text_value, container_field_two_random_text_value

    first_generated_container_field_one_random_text_value, first_generated_container_field_two_random_text_value = generate_update_container_fields_text_values()
    second_table_removed_container_fields_without_data_validation: Dict[str, Optional[Any]] = second_table.remove_multiple_fields(
        key_value=TEST_ACCOUNT_ID, data_validation=False, removers={
            'fieldOne': FieldRemover(field_path='container.nestedFieldOne'),
            'fieldTwo': FieldRemover(field_path='container.nestedFieldTwo')
        }
    )
    self.assertEqual({
        'fieldOne': (
            first_generated_container_field_one_random_text_value if is_caching is not True else
            {'value': first_generated_container_field_one_random_text_value, 'fromCache': False}
        ),
        'fieldTwo': (
            first_generated_container_field_two_random_text_value if is_caching is not True else
            {'value': first_generated_container_field_two_random_text_value, 'fromCache': False}
        )
    }, second_table_removed_container_fields_without_data_validation)

    if is_caching is True:
        second_table.clear_cached_data()

    second_generated_container_field_one_random_text_value, second_generated_container_field_two_random_text_value = generate_update_container_fields_text_values()
    second_table_removed_container_fields_with_data_validation: Dict[str, Optional[int]] = second_table.remove_multiple_fields(
        key_value=TEST_ACCOUNT_ID, data_validation=True, removers={
            'fieldOne': FieldRemover(field_path='container.nestedFieldOne'),
            'fieldTwo': FieldRemover(field_path='container.nestedFieldTwo')
        }
    )
    self.assertEqual({
        'fieldOne': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        ),
        'fieldTwo': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        )
    }, second_table_removed_container_fields_with_data_validation)


def test_grouped_remove_multiple_fields(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    def generate_update_container_fields_text_values() -> Tuple[str, str]:
        container_field_one_random_text_value: str = f"container_fieldOne_randomTextValue_{uuid4()}"
        container_field_two_random_text_value: str = f"container_fieldTwo_randomTextValue_{uuid4()}"
        first_table_container_fields_update_success: bool = first_table.update_multiple_fields(
            key_value=TEST_ACCOUNT_ID, setters=[
                FieldSetter(field_path='container.nestedFieldOne', value_to_set=container_field_one_random_text_value),
                FieldSetter(field_path='container.nestedFieldTwo', value_to_set=container_field_two_random_text_value)
            ]
        )
        self.assertTrue(first_table_container_fields_update_success)
        if is_caching is True:
            self.assertTrue(first_table.commit_operations())
            first_table.clear_cached_data()
        return container_field_one_random_text_value, container_field_two_random_text_value

    first_generated_container_field_one_random_text_value, first_generated_container_field_two_random_text_value = generate_update_container_fields_text_values()
    second_table_removed_container_fields_without_data_validation: Dict[str, Optional[Any]] = second_table.grouped_remove_multiple_fields(
        key_value=TEST_ACCOUNT_ID, data_validation=False, removers={
            'fieldOne': FieldRemover(field_path='container.nestedFieldOne'),
            'fieldTwo': FieldRemover(field_path='container.nestedFieldTwo')
        }
    )
    self.assertEqual({
        'fieldOne': (
            first_generated_container_field_one_random_text_value if is_caching is not True else
            {'value': first_generated_container_field_one_random_text_value, 'fromCache': False}
        ),
        'fieldTwo': (
            first_generated_container_field_two_random_text_value if is_caching is not True else
            {'value': first_generated_container_field_two_random_text_value, 'fromCache': False}
        )
    }, second_table_removed_container_fields_without_data_validation)

    if is_caching is True:
        second_table.clear_cached_data()

    second_generated_container_field_one_random_text_value, second_generated_container_field_two_random_text_value = generate_update_container_fields_text_values()
    second_table_removed_container_fields_with_data_validation: Dict[str, Optional[int]] = second_table.grouped_remove_multiple_fields(
        key_value=TEST_ACCOUNT_ID, data_validation=True, removers={
            'fieldOne': FieldRemover(field_path='container.nestedFieldOne'),
            'fieldTwo': FieldRemover(field_path='container.nestedFieldTwo')
        }
    )
    self.assertEqual({
        'fieldOne': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        ),
        'fieldTwo': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        )
    }, second_table_removed_container_fields_with_data_validation)


def test_update_field_return_old(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    simple_field_random_text_value: str = f"simpleField_randomTextValue_{uuid4()}"
    first_table_simple_field_update_success: bool = first_table.update_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', value_to_set=simple_field_random_text_value
    )
    self.assertTrue(first_table_simple_field_update_success)
    if is_caching is True:
        self.assertTrue(first_table.commit_operations())
        first_table.clear_cached_data()

    second_table_retrieved_simple_field_without_data_validation: Optional[Any] = second_table.get_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', data_validation=False
    )
    self.assertEqual((
        simple_field_random_text_value if is_caching is not True else
        {'value': simple_field_random_text_value, 'fromCache': False}
    ), second_table_retrieved_simple_field_without_data_validation)

    if is_caching is True:
        second_table.clear_cached_data()

    simple_field_dummy_int_value: int = 42
    second_table_simple_field_update_success, second_table_retrieved_old_simple_field_with_data_validation = second_table.update_field_return_old(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', value_to_set=simple_field_dummy_int_value
    )
    second_table_retrieved_old_simple_field: Optional[int]
    self.assertTrue(second_table_simple_field_update_success)
    self.assertEqual((
        None if is_caching is not True else
        {'value': None, 'fromCache': False}
    ), second_table_retrieved_old_simple_field_with_data_validation)


def test_update_multiple_fields_return_old(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    container_field_one_random_text_value: str = f"container_fieldOne_randomTextValue_{uuid4()}"
    container_field_two_random_text_value: str = f"container_fieldTwo_randomTextValue_{uuid4()}"
    first_table_container_fields_update_success: bool = first_table.update_multiple_fields(
        key_value=TEST_ACCOUNT_ID, setters=[
            FieldSetter(field_path='container.nestedFieldOne', value_to_set=container_field_one_random_text_value),
            FieldSetter(field_path='container.nestedFieldTwo', value_to_set=container_field_two_random_text_value)
        ]
    )
    self.assertTrue(first_table_container_fields_update_success)
    if is_caching is True:
        self.assertTrue(first_table.commit_operations())
        first_table.clear_cached_data()

    second_table_retrieved_container_fields_without_data_validation: Dict[str, Optional[Any]] = second_table.get_field(
        key_value=TEST_ACCOUNT_ID, field_path='container.(nestedFieldOne, nestedFieldTwo)', data_validation=False
    )
    self.assertEqual({
        'nestedFieldOne': (
            container_field_one_random_text_value if is_caching is not True else
            {'value': container_field_one_random_text_value, 'fromCache': False}
        ),
        'nestedFieldTwo': (
            container_field_two_random_text_value if is_caching is not True else
            {'value': container_field_two_random_text_value, 'fromCache': False}
        )
    }, second_table_retrieved_container_fields_without_data_validation)

    if is_caching is True:
        second_table.clear_cached_data()

    second_table_container_fields_update_success, second_table_retrieved_container_fields_with_data_validation = second_table.update_multiple_fields_return_old(
        key_value=TEST_ACCOUNT_ID, data_validation=True, setters={
            'fieldOne': FieldSetter(field_path='container.nestedFieldOne', value_to_set=42),
            'fieldTwo': FieldSetter(field_path='container.nestedFieldTwo', value_to_set=42)
        }
    )
    second_table_retrieved_container_fields: Dict[str, Optional[int]]
    self.assertTrue(second_table_container_fields_update_success)
    self.assertEqual({
        'fieldOne': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        ),
        'fieldTwo': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        )
    }, second_table_retrieved_container_fields_with_data_validation)


def test_query_field(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    simple_field_random_text_value: str = f"simpleField_randomTextValue_{uuid4()}"
    first_table_simple_field_update_success: bool = first_table.update_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', value_to_set=simple_field_random_text_value
    )
    self.assertTrue(first_table_simple_field_update_success)
    if is_caching is True:
        self.assertTrue(first_table.commit_operations())
        first_table.clear_cached_data()

    second_table_retrieved_simple_field_without_data_validation, query_metadata = second_table.query_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', data_validation=False
    )
    second_table_retrieved_simple_field_without_data_validation: Optional[Any]
    self.assertEqual({TEST_ACCOUNT_ID: (
        simple_field_random_text_value if is_caching is not True else
        {'value': simple_field_random_text_value, 'fromCache': False}
    )}, second_table_retrieved_simple_field_without_data_validation)
    
    if is_caching is True:
        second_table.clear_cached_data()

    second_table_retrieved_simple_field_with_data_validation, query_metadata = second_table.query_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', data_validation=True
    )
    second_table_retrieved_simple_field_with_data_validation: Optional[int]
    self.assertEqual({TEST_ACCOUNT_ID: (
        None if is_caching is not True else
        {'value': None, 'fromCache': False}
    )}, second_table_retrieved_simple_field_with_data_validation)


def test_query_field_multi_selectors(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    container_field_one_random_text_value: str = f"container_fieldOne_randomTextValue_{uuid4()}"
    container_field_two_random_text_value: str = f"container_fieldTwo_randomTextValue_{uuid4()}"
    first_table_container_fields_update_success: bool = first_table.update_multiple_fields(
        key_value=TEST_ACCOUNT_ID, setters=[
            FieldSetter(field_path='container.nestedFieldOne', value_to_set=container_field_one_random_text_value),
            FieldSetter(field_path='container.nestedFieldTwo', value_to_set=container_field_two_random_text_value)
        ]
    )
    self.assertTrue(first_table_container_fields_update_success)
    if is_caching is True:
        self.assertTrue(first_table.commit_operations())
        first_table.clear_cached_data()

    second_table_retrieved_container_fields_without_data_validation, query_metadata = second_table.query_field(
        key_value=TEST_ACCOUNT_ID, field_path='container.(nestedFieldOne, nestedFieldTwo)', data_validation=False
    )
    second_table_retrieved_container_fields_without_data_validation: Dict[str, Optional[Any]]
    self.assertEqual({TEST_ACCOUNT_ID: {
        'nestedFieldOne': (
            container_field_one_random_text_value if is_caching is not True else
            {'value': container_field_one_random_text_value, 'fromCache': False}
        ),
        'nestedFieldTwo': (
            container_field_two_random_text_value if is_caching is not True else
            {'value': container_field_two_random_text_value, 'fromCache': False}
        )
    }}, second_table_retrieved_container_fields_without_data_validation)
    
    if is_caching is True:
        second_table.clear_cached_data()

    second_table_retrieved_container_fields_with_data_validation, query_metadata = second_table.query_field(
        key_value=TEST_ACCOUNT_ID, field_path='container.(nestedFieldOne, nestedFieldTwo)', data_validation=True
    )
    second_table_retrieved_container_fields_with_data_validation: Dict[str, Optional[int]]
    self.assertEqual({TEST_ACCOUNT_ID: {
        'nestedFieldOne': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        ),
        'nestedFieldTwo': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        )
    }}, second_table_retrieved_container_fields_with_data_validation)


def test_query_multiple_fields(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    container_field_one_random_text_value: str = f"container_fieldOne_randomTextValue_{uuid4()}"
    container_field_two_random_text_value: str = f"container_fieldTwo_randomTextValue_{uuid4()}"
    first_table_container_fields_update_success: bool = first_table.update_multiple_fields(
        key_value=TEST_ACCOUNT_ID, setters=[
            FieldSetter(field_path='container.nestedFieldOne', value_to_set=container_field_one_random_text_value),
            FieldSetter(field_path='container.nestedFieldTwo', value_to_set=container_field_two_random_text_value)
        ]
    )
    self.assertTrue(first_table_container_fields_update_success)

    if is_caching is True:
        self.assertTrue(first_table.commit_operations())
        first_table.clear_cached_data()

    second_table_retrieved_container_fields_without_data_validation, query_metadata = second_table.query_multiple_fields(
        key_value=TEST_ACCOUNT_ID, data_validation=False, getters={
            'one': FieldGetter(field_path='container.nestedFieldOne'),
            'two': FieldGetter(field_path='container.nestedFieldTwo'),
        }
    )
    second_table_retrieved_container_fields_without_data_validation: Dict[str, Optional[Any]]
    self.assertEqual({TEST_ACCOUNT_ID: {
        'one': (
            container_field_one_random_text_value if is_caching is not True else
            {'value': container_field_one_random_text_value, 'fromCache': False}
        ),
        'two': (
            container_field_two_random_text_value if is_caching is not True else
            {'value': container_field_two_random_text_value, 'fromCache': False}
        )
    }}, second_table_retrieved_container_fields_without_data_validation)

    if is_caching is True:
        second_table.clear_cached_data()

    second_table_retrieved_container_fields_with_data_validation, query_metadata = second_table.query_multiple_fields(
        key_value=TEST_ACCOUNT_ID, data_validation=True, getters={
            'one': FieldGetter(field_path='container.nestedFieldOne'),
            'two': FieldGetter(field_path='container.nestedFieldTwo'),
        }
    )
    second_table_retrieved_container_fields_with_data_validation: Dict[str, Optional[int]]
    self.assertEqual({TEST_ACCOUNT_ID: {
        'one': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        ),
        'two': (
            None if is_caching is not True else
            {'value': None, 'fromCache': False}
        )
    }}, second_table_retrieved_container_fields_with_data_validation)


def test_remove_record(
        self: unittest.TestCase,
        first_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        second_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        is_caching: bool, primary_key_name: str
):
    random_record_id: str = f"recordId_{uuid4()}"
    simple_field_random_text_value: str = f"simpleField_randomTextValue_{uuid4()}"

    first_table_put_record_success: bool = first_table.put_record(
        record_dict_data={primary_key_name: random_record_id, 'simpleField': simple_field_random_text_value}
    )
    self.assertTrue(first_table_put_record_success)
    if is_caching is True:
        self.assertTrue(first_table.commit_operations())
        first_table.clear_cached_data()

    second_table_retrieved_simple_field_without_data_validation: Optional[Any] = second_table.get_field(
        key_value=random_record_id, field_path='simpleField', data_validation=False
    )
    self.assertEqual((
        simple_field_random_text_value if is_caching is not True else
        {'value': simple_field_random_text_value, 'fromCache': False}
    ), second_table_retrieved_simple_field_without_data_validation)

    if is_caching is True:
        second_table.clear_cached_data()

    second_table_removed_record_data_with_data_validation: Optional[dict] = second_table.remove_record(
        indexes_keys_selectors={primary_key_name: random_record_id}, data_validation=True
    )
    # The only valid field from the removed_record should be the primary_key field. Since the simpleField
    # will not be of a type matching the model, its key should not be included in the removed_record_data.
    self.assertEqual((
        {primary_key_name: random_record_id} if is_caching is not True else
        {'value': {primary_key_name: random_record_id}, 'fromCache': False}
    ), second_table_removed_record_data_with_data_validation)



import unittest
from typing import Set, Optional, Dict, List
from uuid import uuid4

from StructNoSQL import BaseField, MapModel, TableDataModel
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    nestedDictDictStructure = BaseField(field_type=Dict[str, Dict[str, bool]], required=False, key_name='itemKey')
    # nestedDictListStructure = BaseField(field_type=Dict[str, List[str]], required=False)
    # nestedDictSetStructure = BaseField(field_type=Dict[str, Set[str]], required=False)


class TestsNestedStructuresInsideStructureValues(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_nested_dict_dict_structure(self):
        random_parent_key = f"parentKey_{uuid4()}"
        random_child_key = f"childKey_{uuid4()}"

        keys_fields_switch = list(self.users_table.fields_switch.keys())
        self.assertIn('nestedDictDictStructure.{{itemKey}}.{{itemKeyChild}}', keys_fields_switch)

        update_success = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='nestedDictDictStructure.{{itemKey}}.{{itemKeyChild}}',
            query_kwargs={'itemKey': random_parent_key, 'itemKeyChild': random_child_key},
            value_to_set=True
        )
        self.assertTrue(update_success)

        retrieved_item = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='nestedDictDictStructure.{{itemKey}}',
            query_kwargs={'itemKey': random_parent_key}
        )
        self.assertEqual(retrieved_item, {'itemKeyChild': True})

        removed_item = self.users_table.remove_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='nestedDictDictStructure.{{itemKey}}',
            query_kwargs={'itemKey': random_parent_key}
        )
        self.assertEqual(removed_item, {'itemKeyChild': True})

        retrieved_expected_none_item = self.users_table.get_field(
            TEST_ACCOUNT_ID,
            field_path='nestedDictDictStructure.{{itemKey}}',
            query_kwargs={'itemKey': random_parent_key}
        )
        self.assertIsNone(retrieved_expected_none_item)

    def test_nested_dict_list_structure(self):
        # todo: implement
        pass

    def test_nested_dict_set_structure(self):
        # todo: implement
        pass


if __name__ == '__main__':
    unittest.main()



from typing import List, Dict, Any, Optional, Tuple

from StructNoSQL.clients_middlewares.dynamodb.backend.dynamodb_core import Response

from StructNoSQL.models import DatabasePathElement, FieldPathSetter
from pydantic import BaseModel

from StructNoSQL.servers_middlewares.external_dynamodb_api.low_level_table_client import DynamoDBLowLevelTableClient
from StructNoSQL.servers_middlewares.external_dynamodb_api.models import FieldPathElementItemModel


def putRecord(table: DynamoDBLowLevelTableClient, data: dict) -> Tuple[bool, Optional[dict], dict]:
    class RequestDataModel(BaseModel):
        recordItemData: dict

    request_data = RequestDataModel(**data)
    success: bool = table.dynamodb_client.put_record(
        item_dict=request_data.recordItemData
    )
    return success, None, {}


def deleteRecord(table: DynamoDBLowLevelTableClient, data: dict) -> Tuple[bool, Optional[dict], dict]:
    class RequestDataModel(BaseModel):
        indexesKeysSelectors: Dict[str, Any]

    request_data = RequestDataModel(**data)
    record_deletion_success: bool = table.dynamodb_client.delete_record(
        indexes_keys_selectors=request_data.indexesKeysSelectors
    )
    return record_deletion_success, None, {}


def removeRecord(table: DynamoDBLowLevelTableClient, data: dict) -> Tuple[bool, Optional[dict], dict]:
    class RequestDataModel(BaseModel):
        indexesKeysSelectors: Dict[str, Any]

    request_data = RequestDataModel(**data)
    removed_record_data: Optional[dict] = table.dynamodb_client.remove_record(
        indexes_keys_selectors=request_data.indexesKeysSelectors
    )
    return (True, removed_record_data, {}) if removed_record_data is not None else (False, {}, {})


def queryItemsByKey(table: DynamoDBLowLevelTableClient, data: dict) -> Tuple[bool, Optional[dict], dict]:
    class RequestDataModel(BaseModel):
        keyValue: str
        fieldsPathElements: List[List[FieldPathElementItemModel]]
        paginationRecordsLimit: Optional[int] = None
        filterExpression: Optional[Any] = None

    request_data = RequestDataModel(**data)
    fields_path_elements: List[List[DatabasePathElement]] = [
        [
            DatabasePathElement(
                element_key=item.elementKey,
                default_type=item.defaultType,
                custom_default_value=item.customDefaultValue
            )
            for item in path_elements
        ]
        for path_elements in request_data.fieldsPathElements
    ]
    records_attributes, query_metadata = table.dynamodb_client.query_items_by_key(
        index_name=table.primary_index_name, key_value=request_data.keyValue, fields_path_elements=fields_path_elements,
        pagination_records_limit=request_data.paginationRecordsLimit, filter_expression=request_data.filterExpression
    )
    return True, {'data': records_attributes, 'metadata': query_metadata.serialize()}, {}


def getOrQuerySingleItem(table: DynamoDBLowLevelTableClient, data: dict) -> Tuple[bool, Optional[dict], dict]:
    class RequestDataModel(BaseModel):
        keyValue: str
        fieldsPathElements: List[List[FieldPathElementItemModel]]

    request_data = RequestDataModel(**data)
    fields_path_elements: List[List[DatabasePathElement]] = [
        [
            DatabasePathElement(
                element_key=item.elementKey,
                default_type=item.defaultType,
                custom_default_value=item.customDefaultValue
            )
            for item in path_elements_container
        ]
        for path_elements_container in request_data.fieldsPathElements
    ]
    response_data = table.dynamodb_client.get_or_query_single_item(
        index_name=table.primary_index_name, key_value=request_data.keyValue,
        fields_path_elements=fields_path_elements
    )
    return True, response_data, {}


def getSingleValueInPathTarget(table: DynamoDBLowLevelTableClient, data: dict) -> Tuple[bool, Optional[dict], dict]:
    class RequestDataModel(BaseModel):
        keyValue: str
        fieldPathElements: List[FieldPathElementItemModel]

    request_data = RequestDataModel(**data)
    field_path_elements: List[DatabasePathElement] = [
        DatabasePathElement(
            element_key=item.elementKey,
            default_type=item.defaultType,
            custom_default_value=item.customDefaultValue
        )
        for item in request_data.fieldPathElements
    ]
    response_data = table.dynamodb_client.get_value_in_path_target(
        index_name=table.primary_index_name, key_value=request_data.keyValue,
        field_path_elements=field_path_elements
    )
    return True, response_data, {}


def getValuesInMultiplePathTarget(table: DynamoDBLowLevelTableClient, data: dict) -> Tuple[bool, Optional[dict], dict]:
    class RequestDataModel(BaseModel):
        keyValue: str
        fieldsPathElements: Dict[str, List[FieldPathElementItemModel]]

    request_data = RequestDataModel(**data)
    fields_path_elements: Dict[str, List[DatabasePathElement]] = {
        key: [
            DatabasePathElement(
                element_key=item.elementKey,
                default_type=item.defaultType,
                custom_default_value=item.customDefaultValue
            )
            for item in path_elements_container
        ]
        for key, path_elements_container in request_data.fieldsPathElements.items()
    }
    response_data = table.dynamodb_client.get_values_in_multiple_path_target(
        index_name=table.primary_index_name, key_value=request_data.keyValue,
        fields_path_elements=fields_path_elements, metadata=False
    )
    return (True, response_data, {}) if response_data is not None else (False, {}, {})

def setUpdateMultipleDataElementsToMap(table: DynamoDBLowLevelTableClient, data: dict) -> Tuple[bool, Optional[dict], dict]:
    class RequestDataModel(BaseModel):
        keyValue: str
        class SetterItemModel(BaseModel):
            valueToSet: Any
            fieldPathElements: List[FieldPathElementItemModel]
        setters: List[SetterItemModel]
        returnOldValues: bool = False

    request_data = RequestDataModel(**data)
    setters: List[FieldPathSetter] = [
        FieldPathSetter(
            value_to_set=setter_item.valueToSet,
            field_path_elements=[
                DatabasePathElement(
                    element_key=item.elementKey,
                    default_type=item.defaultType,
                    custom_default_value=item.customDefaultValue
                )
                for item in setter_item.fieldPathElements
            ]
        )
        for setter_item in request_data.setters
    ]
    response: Optional[Response] = table.dynamodb_client.set_update_multiple_data_elements_to_map(
        index_name=table.primary_index_name, key_value=request_data.keyValue,
        setters=setters, return_old_values=request_data.returnOldValues
    )
    if response is None:
        return False, None, {}

    from StructNoSQL.clients_middlewares.dynamodb.backend.dynamodb_utils import DynamoDBUtils
    response_attributes: Optional[dict] = (
        DynamoDBUtils.dynamodb_to_python_higher_level(response.attributes)
        if response.attributes is not None else None
    )
    return True, response_attributes, {}

def setUpdateMultipleDataElementsToMapWithDefaultInitialization(table: DynamoDBLowLevelTableClient, data: dict) -> Tuple[bool, Optional[dict], dict]:
    class RequestDataModel(BaseModel):
        keyValue: str
        fieldPathElements: List[FieldPathElementItemModel]
        value: Any
        returnOldValue: bool = False

    request_data = RequestDataModel(**data)
    field_path_elements: List[DatabasePathElement] = [
        DatabasePathElement(
            element_key=item.elementKey,
            default_type=item.defaultType,
            custom_default_value=item.customDefaultValue
        )
        for item in request_data.fieldPathElements
    ]
    response: Optional[Response] = table.dynamodb_client.set_update_data_element_to_map_with_default_initialization(
        index_name=table.primary_index_name, key_value=request_data.keyValue, field_path_elements=field_path_elements,
        return_old_value=request_data.returnOldValue, value=request_data.value,
    )
    if response is None:
        return False, None, {}

    from StructNoSQL.clients_middlewares.dynamodb.backend.dynamodb_utils import DynamoDBUtils
    response_attributes: Optional[Dict[str, Any]] = (
        DynamoDBUtils.dynamodb_to_python_higher_level(response.attributes)
        if response.attributes is not None else None
    )
    return True, response_attributes, {}

def removeDataElementsFromMap(table: DynamoDBLowLevelTableClient, data: dict) -> Tuple[bool, Optional[dict], dict]:
    class RequestDataModel(BaseModel):
        keyValue: str
        fieldsPathElements: List[List[FieldPathElementItemModel]]

    request_data = RequestDataModel(**data)
    fields_path_elements: List[List[DatabasePathElement]] = [
        [
            DatabasePathElement(
                element_key=item.elementKey,
                default_type=item.defaultType,
                custom_default_value=item.customDefaultValue
            )
            for item in path_elements_container
        ]
        for path_elements_container in request_data.fieldsPathElements
    ]
    response_data = table.dynamodb_client.remove_data_elements_from_map(
        index_name=table.primary_index_name, key_value=request_data.keyValue,
        targets_path_elements=fields_path_elements,
        retrieve_removed_elements=True
    )
    return (True, response_data, {}) if response_data is not None else (False, {}, {})

def deleteDataElementsFromMap(table: DynamoDBLowLevelTableClient, data: dict) -> Tuple[bool, Optional[dict], dict]:
    class RequestDataModel(BaseModel):
        keyValue: str
        fieldsPathElements: List[List[FieldPathElementItemModel]]

    request_data = RequestDataModel(**data)
    fields_path_elements: List[List[DatabasePathElement]] = [
        [
            DatabasePathElement(
                element_key=item.elementKey,
                default_type=item.defaultType,
                custom_default_value=item.customDefaultValue
            )
            for item in path_elements_container
        ]
        for path_elements_container in request_data.fieldsPathElements
    ]
    response_data = table.dynamodb_client.remove_data_elements_from_map(
        index_name=table.primary_index_name, key_value=request_data.keyValue,
        targets_path_elements=fields_path_elements,
        retrieve_removed_elements=False
    )
    return True if response_data is not None else False, {}, {}



from typing import Optional, Any, List

from pydantic import ValidationError

from StructNoSQL import DynamoDBTableConnectors, GlobalSecondaryIndex, PrimaryIndex
from StructNoSQL.clients_middlewares.dynamodb.backend.dynamodb_core import DynamoDbCoreAdapter
from StructNoSQL.servers_middlewares.external_dynamodb_api.low_level_table_client import DynamoDBLowLevelTableClient
from StructNoSQL.servers_middlewares.external_dynamodb_api.parser import StructNoSQLParser


class ExternalDynamoDBApiExecutor(DynamoDBTableConnectors):
    def __init__(
            self, table_name: str, region_name: str, primary_index: PrimaryIndex,
            billing_mode: str = DynamoDbCoreAdapter.PAY_PER_REQUEST,
            global_secondary_indexes: List[GlobalSecondaryIndex] = None,
            auto_create_table: bool = True
    ):
        self.table = DynamoDBLowLevelTableClient(
            table_name=table_name, region_name=region_name,
            primary_index=primary_index, global_secondary_indexes=global_secondary_indexes,
            billing_mode=billing_mode, auto_create_table=auto_create_table
        )

    def execute(self, data: dict) -> dict:
        handler: Optional[Any] = StructNoSQLParser.parse(data=data)
        if handler is None:
            return {'success': False, 'errorKey': 'operationNotSupported'}

        try:
            success, data, kwargs = handler(self.table, data)
            return {'success': success, 'data': data, **(kwargs or {})}
        except ValidationError as e:
            return {'success': False, 'errorKey': 'validationError', 'exception': str(e)}



from json import JSONDecodeError
from typing import Optional, Any

from flask import Flask
from flask import request, jsonify
from tests.components.playground_table_servers import InoftVocalEngineExternalDynamoDBApiExecutor


app = Flask(__name__)


def get_access_token_from_args_or_payload() -> Optional[str]:
    arg_access_token: Optional[str] = request.args.get('accessToken', None)
    if arg_access_token is not None:
        return arg_access_token

    try:
        request_payload: Optional[dict] = request.get_json()
        if request_payload is not None and isinstance(request_payload, dict):
            payload_access_token: Optional[str] = request_payload.get('accessToken', None)
            if payload_access_token is not None and isinstance(payload_access_token, str):
                return payload_access_token
    except JSONDecodeError as e:
        pass

    return None


@app.route("/api/v1/<account_id>/<project_id>/database-client", methods=['POST'])
def database_client(account_id: str, project_id: str):
    access_token_id: Optional[str] = get_access_token_from_args_or_payload()
    if access_token_id is None:
        return jsonify(success=False, errorKey='invalidOrMissingAccessToken')

    if access_token_id != 'daisybelle':
        return jsonify(success=False, errorKey='invalidAccessToken')

    try:
        request_payload: Any = request.get_json()
        if not isinstance(request_payload, dict):
            return jsonify(success=False, errorKey='expectedDictPayload')

        return InoftVocalEngineExternalDynamoDBApiExecutor().execute(request_payload)

    except JSONDecodeError as e:
        return jsonify(success=False, errorKey='payloadInvalidJSONFormat', error=str(e))


if __name__ == '__main__':
    app.run()



import unittest
from typing import List, Optional
from uuid import uuid4

from StructNoSQL import FieldGetter, FieldSetter, FieldRemover
from StructNoSQL.exceptions import FieldTargetNotFoundException
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID, TEST_PROJECT_ID, TEST_ACCOUNT_EMAIL, TEST_ACCOUNT_USERNAME
from tests.users_table_model import UsersTableModel


class TestTableOperations(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=UsersTableModel)

    def test_get_name_of_one_project(self):
        response_data: Optional[str] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='projects.{{projectId}}.projectName',
            query_kwargs={'projectId': TEST_PROJECT_ID}
        )
        self.assertIn("test", response_data)

    def test_update_project_name(self):
        success = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='projects.{{projectId}}.projectName',
            query_kwargs={'projectId': TEST_PROJECT_ID},
            value_to_set="test2"
        )
        self.assertTrue(success)

        response_data: Optional[str] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='projects.{{projectId}}.projectName',
            query_kwargs={'projectId': TEST_PROJECT_ID}
        )
        self.assertEqual(response_data, "test2")

    def test_update_entire_project_model(self):
        success = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='projects.{{projectId}}',
            value_to_set={'projectName': "test3"},
            query_kwargs={'projectId': TEST_PROJECT_ID}
        )

        # todo: allow to set the item of a dict (currently, when doing a query on the projects object,
        #  we will perform an operation of the project map, and not on an individual project item).
        self.assertTrue(success)

        response_data: Optional[str] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='projects.{{projectId}}.projectName',
            query_kwargs={'projectId': TEST_PROJECT_ID}
        )
        self.assertEqual(response_data, "test3")

    def test_update_entire_project_model_with_missing_project_name(self):
        success = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='projects.{{projectId}}',
            value_to_set={},
            query_kwargs={'projectId': TEST_PROJECT_ID}
        )
        self.assertFalse(success)

    def test_update_entire_project_model_with_invalid_data(self):
        success = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='projects.{{projectId}}',
            value_to_set={'invalidProjectName': "test4"},
            query_kwargs={'projectId': TEST_PROJECT_ID}
        )
        self.assertFalse(success)

        try:
            response_data: Optional[str] = self.users_table.get_field(
                key_value=TEST_ACCOUNT_ID,
                field_path='projects.{{projectId}}.invalidProjectName',
                query_kwargs={'projectId': TEST_PROJECT_ID}
            )
            # If we do not get an error while trying to access an invalid field in the
            # get_field function, then we failed the test.
            self.fail()
        except FieldTargetNotFoundException as e:
            print(e)

    def test_multi_types_field(self):
        success_str: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path="multiTypes",
            value_to_set="yolo"
        )
        self.assertTrue(success_str)

        success_none: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path="multiTypes",
            value_to_set=None
        )
        self.assertTrue(success_none)

        success_bool: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path="multiTypes",
            value_to_set=True
        )
        self.assertFalse(success_bool)

    def test_basic_get_multiple_fields_values_in_single_query(self):
        response_data: Optional[dict] = self.users_table.get_multiple_fields(
            key_value=TEST_ACCOUNT_ID, getters={
                'theAccountId': FieldGetter(field_path='accountId'),
                'theProjects': FieldGetter(field_path='projects')
            }
        )
        self.assertIsNotNone(response_data)
        self.assertEqual(response_data.get('theAccountId', None), TEST_ACCOUNT_ID)

    def test_sophisticated_get_multiple_fields_in_single_query(self):
        response_data: Optional[dict] = self.users_table.get_multiple_fields(
            key_value=TEST_ACCOUNT_ID, getters={
                'theAccountId': FieldGetter(field_path='accountId'),
                'theProjectName': FieldGetter(
                    field_path='projects.{{projectId}}.projectName',
                    query_kwargs={'projectId': TEST_PROJECT_ID}
                )
            }
        )
        self.assertIsNotNone(response_data)
        self.assertEqual(response_data.get('theAccountId', None), TEST_ACCOUNT_ID)
        self.assertIn("test", response_data.get('theProjectName', ""))

    def test_basic_update_multiple_fields_in_single_query(self):
        success: bool = self.users_table.update_multiple_fields(
            key_value=TEST_ACCOUNT_ID,  setters=[
                FieldSetter(field_path='number1', value_to_set=42),
                FieldSetter(field_path='string1', value_to_set="Quarante-deux")
            ]
        )
        self.assertTrue(success)

    def test_sophisticated_update_multiple_fields_in_single_query(self):
        success: bool = self.users_table.update_multiple_fields(
            key_value=TEST_ACCOUNT_ID, setters=[
                FieldSetter(field_path='number1', value_to_set=42),
                FieldSetter(
                    field_path='projects.{{projectId}}.projectName',
                    query_kwargs={'projectId': TEST_PROJECT_ID},
                    value_to_set="test5",
                )
            ]
        )
        self.assertTrue(success)

        project_name_data: Optional[str] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='projects.{{projectId}}.projectName',
            query_kwargs={'projectId': TEST_PROJECT_ID},
        )
        self.assertEqual(project_name_data, "test5")

    def test_set_and_get_float_in_field_value(self):
        source_float_value = 10.42021023492

        set_float_value_success = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='floatTest',
            value_to_set=source_float_value
        )
        self.assertTrue(set_float_value_success)

        retrieved_float_value: Optional[float] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='floatTest'
        )
        self.assertEqual(retrieved_float_value, source_float_value)

    def test_remove_basic_item_from_path_target(self):
        success_field_set: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path="fieldToDelete",
            value_to_set="yolo mon ami !"
        )
        self.assertTrue(success_field_set)

        success_field_remove: bool = self.users_table.delete_field(
            key_value=TEST_ACCOUNT_ID,
            field_path="fieldToDelete"
        )
        self.assertTrue(success_field_remove)

        retrieved_field_data: Optional[float] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path="fieldToDelete"
        )
        self.assertIsNone(retrieved_field_data)

    def test_remove_sophisticated_item_from_path_target(self):
        query_kwargs = {"id": "sampleId"}

        success_field_set = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='sophisticatedRemoval.{{id}}.nestedVariable',
            query_kwargs=query_kwargs,
            value_to_set="Soooooo, does it works ? :)"
        )
        self.assertTrue(success_field_set)

        success_field_remove = self.users_table.delete_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='sophisticatedRemoval.{{id}}.nestedVariable',
            query_kwargs=query_kwargs
        )
        self.assertTrue(success_field_remove)

        retrieved_field_data = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='sophisticatedRemoval.{{id}}.nestedVariable',
            query_kwargs=query_kwargs
        )
        self.assertIsNone(retrieved_field_data)

    def test_remove_multiple_basic_and_sophisticated_items_from_path_target(self):
        query_kwargs = {"id": "sampleIdTwo"}

        success_fields_set: bool = self.users_table.update_multiple_fields(
            key_value=TEST_ACCOUNT_ID, setters=[
                FieldSetter(
                    field_path='fieldToDelete',
                    value_to_set="multipleBasicAndSophisticatedRemoval"
                ),
                FieldSetter(
                    field_path='sophisticatedRemoval.{{id}}.nestedVariable',
                    query_kwargs=query_kwargs, value_to_set="nestedDude"
                )
            ]
        )
        self.assertTrue(success_fields_set)

        deletion_response = self.users_table.delete_multiple_fields(
            key_value=TEST_ACCOUNT_ID, removers={
                'field': FieldRemover(field_path="fieldToDelete"),
                'nested': FieldRemover(field_path="sophisticatedRemoval.{{id}}.nestedVariable", query_kwargs=query_kwargs)
            }
        )
        self.assertTrue(all(deletion_response.values()))

    def test_put_and_delete_records(self):
        success_put_invalid_record = self.users_table.put_record(record_dict_data={"invalidAccountId": "testAccountId", "multiTypes": "testPutRecord"})
        self.assertFalse(success_put_invalid_record)

        success_put_valid_record = self.users_table.put_record(record_dict_data={"accountId": "testAccountId", "multiTypes": "testPutRecord"})
        self.assertTrue(success_put_valid_record)

        success_delete_record_with_index_typo = self.users_table.delete_record(indexes_keys_selectors={"accountIdWithTypo": "testAccountId"})
        self.assertFalse(success_delete_record_with_index_typo)

        success_delete_record_without_typo = self.users_table.delete_record(indexes_keys_selectors={"accountId": "testAccountId"})
        self.assertTrue(success_delete_record_without_typo)

    def test_get_value_from_path_target_by_secondary_index(self):
        account_id: Optional[str] = self.users_table.get_field(
            index_name="email", key_value=TEST_ACCOUNT_EMAIL, field_path="accountId"
        )
        self.assertEqual(account_id, TEST_ACCOUNT_ID)

        account_data: Optional[dict] = self.users_table.get_multiple_fields(
            index_name="email", key_value=TEST_ACCOUNT_EMAIL,
            getters={
                "accountId": FieldGetter(field_path="accountId"),
                "username": FieldGetter(field_path="username")
            }
        )
        self.assertEqual(account_data.get("accountId", None), TEST_ACCOUNT_ID)
        self.assertEqual(account_data.get("username", None), TEST_ACCOUNT_USERNAME)

    def test_set_data_inside_a_map_model_field(self):
        dummy_value = str(uuid4())

        set_update_success: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='testMapModel.sampleText',
            value_to_set=dummy_value
        )
        self.assertEqual(set_update_success, True)

        retrieved_value: Optional[str] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='testMapModel.sampleText'
        )
        self.assertEqual(retrieved_value, dummy_value)

        delete_field_success: bool = self.users_table.delete_field(
            key_value=TEST_ACCOUNT_ID,
            field_path="testMapModel.sampleText"
        )
        self.assertEqual(delete_field_success, True)

    def test_set_dict_item_with_primitive_value(self):
        success_valid: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='testDictWithPrimitiveValue.{{key}}',
            query_kwargs={'key': "one"},
            value_to_set=True
        )
        self.assertTrue(success_valid)

        success_invalid: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='testDictWithPrimitiveValue.{{key}}',
            query_kwargs={'key': "one"},
            value_to_set={'keyOfDictThatShouldNotBeHere': True}
        )
        self.assertFalse(success_invalid)

if __name__ == '__main__':
    unittest.main()



import unittest
from typing import Dict
from StructNoSQL import BaseField, MapModel, TableDataModel, PrimaryIndex, GlobalSecondaryIndex, ActiveSelf, \
    DynamoDBBasicTable


class NestedFieldTableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    class PropertyModel(MapModel):
        name = BaseField(field_type=str)
        childProperties = BaseField(field_type=Dict[str, ActiveSelf], key_name="childPropertyKey{i}", required=False)
    properties = BaseField(field_type=Dict[str, PropertyModel], key_name="propertyKey", required=False)


class Table(DynamoDBBasicTable):
    def __init__(self):
        primary_index = PrimaryIndex(hash_key_name="accountId", hash_key_variable_python_type=str)
        globals_secondary_indexes = [
            GlobalSecondaryIndex(hash_key_name="username", hash_key_variable_python_type=str, projection_type="ALL"),
            GlobalSecondaryIndex(hash_key_name="email", hash_key_variable_python_type=str, projection_type="ALL"),
        ]
        super().__init__(
            table_name="structnosql-playground", region_name="eu-west-2",
            data_model=NestedFieldTableModel(),
            primary_index=primary_index, global_secondary_indexes=globals_secondary_indexes,
            auto_create_table=True
        )


class TestTableOperations(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = Table()

    def test_get_all_projects(self):
        print(self.users_table.fields_switch)
        for field in self.users_table.fields_switch.keys():
            print(field)

        """success_put = self.users_table.put_record({
            'accountId': 'staticId', 'properties': {}
        })
        self.assertTrue(success_put)"""

        success_update = self.users_table.update_field(
            key_value='staticId',
            field_path='properties.{{propertyKey}}.childProperties.{{childPropertyKey0}}.childProperties.{{childPropertyKey1}}.name',
            value_to_set="theGreatCode", query_kwargs={
                'propertyKey': 'staticPropertyKey',
                'childPropertyKey0': 'staticChildPropertyKey0',
                'childPropertyKey1': 'staticChildPropertyKey1'
            }
        )
        self.assertTrue(success_update)


if __name__ == '__main__':
    unittest.main()



import logging
from typing import List, Optional, Any, Dict, Tuple
from json import JSONDecodeError

from pydantic import BaseModel, ValidationError

from StructNoSQL.models import FieldPathSetter, DatabasePathElement, QueryMetadata
import requests


class ExternalDynamoDBApiTableConnectors:
    def __setup_connectors__(self, api_http_endpoint: str, base_payload: Optional[dict] = None):
        self.api_http_endpoint = api_http_endpoint
        self.base_payload = base_payload or {}

    def _base_api_handler(self, payload: dict) -> Optional[dict]:
        response = requests.post(
            url=self.api_http_endpoint,
            json={**payload, **self.base_payload}
        )
        if not response.ok:
            print("Response not ok")
            return None
        else:
            try:
                response_data: Optional[dict] = response.json()
                return response_data if response_data is not None and isinstance(response_data, dict) else None
            except JSONDecodeError as e:
                print(f"JSON decoding error : {e} : {response.text}")
                return None

    def _success_api_handler(self, payload: dict) -> bool:
        response_data: Optional[dict] = self._base_api_handler(payload=payload)
        if response_data is not None:
            success: bool = response_data.get('success', False)
            return success
        return False

    def _data_api_handler(self, payload: dict) -> Optional[Any]:
        response_data: Optional[dict] = self._base_api_handler(payload=payload)
        if response_data is not None:
            success: bool = response_data.get('success', False)
            if success is True:
                return response_data.get('data', None)
            """exception: Optional[str] = response_data.get('exception', None)
            if exception is not None:
                raise Exception(exception)"""
        return None

    def _success_and_data_api_handler(self, payload: dict) -> Tuple[bool, Optional[Any]]:
        response_data: Optional[dict] = self._base_api_handler(payload=payload)
        if response_data is not None:
            success: bool = response_data.get('success', False)
            if success is True:
                return True, response_data.get('data', None)
        return False, None

    def _put_record_request(self, record_item_data: dict) -> bool:
        return self._success_api_handler(payload={
            'operationType': 'putRecord',
            'recordItemData': record_item_data,
        })

    def _delete_record_request(self, indexes_keys_selectors: Dict[str, Any]) -> bool:
        return self._success_api_handler(payload={
            'operationType': 'deleteRecord',
            'indexesKeysSelectors': indexes_keys_selectors,
        })

    def _remove_record_request(self, indexes_keys_selectors: Dict[str, Any]) -> Optional[dict]:
        return self._data_api_handler(payload={
            'operationType': 'removeRecord',
            'indexesKeysSelectors': indexes_keys_selectors,
        })

    def _query_items_by_key(
            self, key_value: str, fields_path_elements: List[List[DatabasePathElement]],
            pagination_records_limit: int, filter_expression: Optional[Any] = None, **additional_kwargs
    ) -> Tuple[Optional[List[dict]], QueryMetadata]:
        serialized_fields_path_elements: List[List[dict]] = [
            [item.serialize() for item in path_elements]
            for path_elements in fields_path_elements
        ]
        response_payload: Optional[dict] = self._data_api_handler(payload={
            'operationType': 'queryItemsByKey',
            'keyValue': key_value,
            'fieldsPathElements': serialized_fields_path_elements,
            'paginationRecordsLimit': pagination_records_limit,
            'filterExpression': filter_expression,
            **additional_kwargs
        })
        if response_payload is not None:
            class ResponseDataModel(BaseModel):
                data: Optional[List[dict]] = None
                class MetadataModel(BaseModel):
                    count: int
                    hasReachedEnd: bool
                    lastEvaluatedKey: Optional[dict]
                metadata: MetadataModel
            try:
                response_data = ResponseDataModel(**response_payload)
                return response_data.data, QueryMetadata(
                    count=response_data.metadata.count,
                    has_reached_end=response_data.metadata.hasReachedEnd,
                    last_evaluated_key=response_data.metadata.lastEvaluatedKey
                )
            except ValidationError as e:
                logging.warning(e)
        return None, QueryMetadata(count=0, has_reached_end=True, last_evaluated_key=None)


    def _set_update_data_element_to_map_with_default_initialization(
            self, key_value: Any, field_path_elements: List[DatabasePathElement], value: Any
    ) -> bool:
        serialized_field_path_elements: List[dict] = [item.serialize() for item in field_path_elements]
        return self._success_api_handler(payload={
            'operationType': 'setUpdateMultipleDataElementsToMapWithDefaultInitialization',
            'keyValue': key_value,
            'fieldPathElements': serialized_field_path_elements,
            'value': value
        })

    def _set_update_data_element_to_map_with_default_initialization_return_old(
            self, key_value: Any, field_path_elements: List[DatabasePathElement], value: Any
    ) -> Tuple[bool, Optional[Any]]:
        serialized_field_path_elements: List[dict] = [item.serialize() for item in field_path_elements]
        return self._success_and_data_api_handler(payload={
            'operationType': 'setUpdateMultipleDataElementsToMapWithDefaultInitialization',
            'keyValue': key_value,
            'fieldPathElements': serialized_field_path_elements,
            'value': value,
            'returnOldValue': True
        })

    def _set_update_multiple_data_elements_to_map(self, key_value: Any, setters: List[FieldPathSetter]) -> bool:
        serialized_setters: List[dict] = [item.serialize() for item in setters]
        return self._success_api_handler(payload={
            'operationType': 'setUpdateMultipleDataElementsToMap',
            'keyValue': key_value,
            'setters': serialized_setters
        })

    def _set_update_multiple_data_elements_to_map_return_old(self, key_value: Any, setters: List[FieldPathSetter]) -> Tuple[bool, Dict[str, Any]]:
        serialized_setters: List[dict] = [item.serialize() for item in setters]
        return self._success_and_data_api_handler(payload={
            'operationType': 'setUpdateMultipleDataElementsToMap',
            'keyValue': key_value,
            'setters': serialized_setters,
            'returnOldValues': True
        })

    def _remove_data_elements_from_map(self, key_value: str, fields_path_elements: List[List[DatabasePathElement]]) -> Optional[Any]:
        serialized_fields_path_elements: List[List[dict]] = [
            [item.serialize() for item in path_elements]
            for path_elements in fields_path_elements
        ]
        return self._data_api_handler(payload={
            'operationType': 'removeDataElementsFromMap',
            'keyValue': key_value,
            'fieldsPathElements': serialized_fields_path_elements
        })

    def _delete_data_elements_from_map(self, key_value: str, fields_path_elements: List[List[DatabasePathElement]]) -> bool:
        serialized_fields_path_elements: List[List[dict]] = [
            [item.serialize() for item in path_elements]
            for path_elements in fields_path_elements
        ]
        return self._success_api_handler(payload={
            'operationType': 'deleteDataElementsFromMap',
            'keyValue': key_value,
            'fieldsPathElements': serialized_fields_path_elements
        })

    def _get_single_value_in_path_target(self, key_value: str, field_path_elements: List[DatabasePathElement]) -> Optional[Any]:
        serialized_field_path_elements: List[dict] = [item.serialize() for item in field_path_elements]
        return self._data_api_handler(payload={
            'operationType': 'getSingleValueInPathTarget',
            'keyValue': key_value,
            'fieldPathElements': serialized_field_path_elements
        })

    def _get_values_in_multiple_path_target(self, key_value: str, fields_path_elements: Dict[str, List[DatabasePathElement]]) -> Optional[Any]:
        serialized_fields_path_elements: Dict[str, List[dict]] = {
            field_key: [item.serialize() for item in field_path_elements_items]
            for field_key, field_path_elements_items in fields_path_elements.items()
        }
        return self._data_api_handler(payload={
            'operationType': 'getValuesInMultiplePathTarget',
            'keyValue': key_value,
            'fieldsPathElements': serialized_fields_path_elements
        })

    def _get_or_query_single_item(self, key_value: str, fields_path_elements: List[List[DatabasePathElement]]) -> Optional[Any]:
        serialized_fields_path_elements: List[List[dict]] = [
            [item.serialize() for item in items_container]
            for items_container in fields_path_elements
        ]
        return self._data_api_handler(payload={
            'operationType': 'getOrQuerySingleItem',
            'keyValue': key_value,
            'fieldsPathElements': serialized_fields_path_elements
        })




import random
import unittest
from uuid import uuid4
from typing import Dict, Any, Union, List, Optional, Generator, Tuple

from StructNoSQL import FieldSetter, FieldGetter, DynamoDBCachingTable, ExternalDynamoDBApiCachingTable, \
    DynamoDBBasicTable, QueryMetadata
from StructNoSQL.clients_middlewares.external_dynamodb_api.external_dynamodb_api_basic_table import ExternalDynamoDBApiBasicTable
from tests.components.playground_table_clients import TEST_ACCOUNT_ID, TEST_ACCOUNT_USERNAME


def test_set_get_fields_with_primary_index(
        self: unittest.TestCase, users_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        primary_key_name: str, is_caching: bool
):
    if is_caching is True:
        users_table.clear_cached_data_and_pending_operations()

    field1_random_value: str = f"field1_{uuid4()}"
    field2_random_value: str = f"field2_{uuid4()}"

    set_update_success: bool = users_table.update_multiple_fields(key_value=TEST_ACCOUNT_ID, setters=[
        FieldSetter(field_path='fieldOne', value_to_set=field1_random_value),
        FieldSetter(field_path='fieldTwo', value_to_set=field2_random_value)
    ])
    self.assertTrue(set_update_success)

    if is_caching is True:
        update_commit_success: bool = users_table.commit_operations()
        self.assertTrue(update_commit_success)

    retrieved_values = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='(fieldOne, fieldTwo)')
    self.assertEqual({
        'fieldOne': (
            field1_random_value if is_caching is not True else
            {'fromCache': True, 'value': field1_random_value}
        ),
        'fieldTwo': (
            field2_random_value if is_caching is not True else
            {'fromCache': True, 'value': field2_random_value}
        )
    }, retrieved_values)

    single_field_not_primary_key, query_metadata = users_table.query_field(key_value=TEST_ACCOUNT_ID, field_path='fieldOne')
    self.assertEqual({TEST_ACCOUNT_ID: (
        field1_random_value if is_caching is not True else
        {'fromCache': True, 'value': field1_random_value}
    )}, single_field_not_primary_key)

    single_field_primary_key, query_metadata = users_table.query_field(key_value=TEST_ACCOUNT_ID, field_path=f'{primary_key_name}')
    self.assertEqual({TEST_ACCOUNT_ID: (
        TEST_ACCOUNT_ID if is_caching is not True else
        {'fromCache': False, 'value': TEST_ACCOUNT_ID}
    )}, single_field_primary_key)

    multiple_fields_without_primary_key, query_metadata = users_table.query_field(
        key_value=TEST_ACCOUNT_ID, field_path='(fieldOne, fieldTwo)'
    )
    self.assertEqual({
        TEST_ACCOUNT_ID: {
            'fieldOne': (
                field1_random_value if is_caching is not True else
                {'fromCache': True, 'value': field1_random_value}
            ),
            'fieldTwo': (
                field2_random_value if is_caching is not True else
                {'fromCache': True, 'value': field2_random_value}
            )
        }
    }, multiple_fields_without_primary_key)

    multiple_fields_with_primary_key, query_metadata = users_table.query_field(
        key_value=TEST_ACCOUNT_ID, field_path=f'({primary_key_name}, fieldOne, fieldTwo)'
    )
    self.assertEqual({
        TEST_ACCOUNT_ID: {
            f'{primary_key_name}': (
                TEST_ACCOUNT_ID if is_caching is not True else
                {'fromCache': True, 'value': TEST_ACCOUNT_ID}
            ),
            'fieldOne': (
                field1_random_value if is_caching is not True else
                {'fromCache': True, 'value': field1_random_value}
            ),
            'fieldTwo': (
                field2_random_value if is_caching is not True else
                {'fromCache': True, 'value': field2_random_value}
            )
        }
    }, multiple_fields_with_primary_key)


def test_set_get_paginated_fields_with_primary_index(
        self: unittest.TestCase, users_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        primary_key_name: str, is_caching: bool
):
    random_type_id: str = f"type_{uuid4()}"

    def make_record() -> str:
        for i_attempt in range(10):
            new_record_id: str = str(uuid4())
            existing_primary_key_value: Optional[Any] = users_table.get_field(
                key_value=new_record_id, field_path=primary_key_name
            )
            no_existing_record_found_for_new_record_id: bool = (
                existing_primary_key_value is None if is_caching is not True else
                existing_primary_key_value['value'] is None
            )
            if no_existing_record_found_for_new_record_id is True:
                put_record_success: bool = users_table.put_record({
                    primary_key_name: new_record_id,
                    'username': 'dummyUsername',
                    'email': 'dummyEmail',
                    'type': random_type_id
                })
                self.assertTrue(put_record_success)
                return new_record_id
        self.fail(msg="Could not generate unique record id after ten attempts")

    created_account_ids: List[str] = [make_record() for _ in range(5)]

    records_paginator: Generator[Tuple[Optional[dict], QueryMetadata], None, None] = users_table.paginated_query_field(
        index_name='type', key_value=random_type_id,
        field_path='accountId', pagination_records_limit=2
    )
    for records_data, query_metadata in records_paginator:
        if records_data is not None:
            for record_item_data in records_data:
                record_deletion_success: bool = users_table.delete_record(indexes_keys_selectors={
                    primary_key_name: record_item_data
                })
                self.assertTrue(record_deletion_success)



import unittest
from time import time
from typing import Optional, Dict
from uuid import uuid4

from StructNoSQL import BaseField, MapModel, FieldGetter, TableDataModel
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    class ItemWithMultipleAttributes(MapModel):
        name = BaseField(field_type=str, required=False)
        value = BaseField(field_type=str, required=False)
        timestamp = BaseField(field_type=(int, float), required=False)
    multiAttributesContainer = BaseField(field_type=Dict[str, ItemWithMultipleAttributes], key_name='itemKey', required=False)


class TestDatabaseFieldsInitialization(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_retrieve_multiple_attributes_from_single_object(self):
        """
        Test both the get_field and the get_multiple_fields function
        """

        randomized_item_key = f"key_{uuid4()}"
        random_name = f"name-{uuid4()}"
        random_value = f"value-{uuid4()}"
        timestamp = time()
        randomized_item = {
            'name': random_name,
            'value': random_value,
            'timestamp': timestamp
        }
        query_kwargs = {'itemKey': randomized_item_key}

        set_success: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='multiAttributesContainer.{{itemKey}}',
            query_kwargs=query_kwargs, value_to_set=randomized_item
        )
        self.assertTrue(set_success)

        retrieved_values: dict = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='multiAttributesContainer.{{itemKey}}.(name, value)',
            query_kwargs=query_kwargs
        )
        self.assertEqual(retrieved_values.get('name'), random_name)
        self.assertEqual(retrieved_values.get('value'), random_value)

        retrieved_values: dict = self.users_table.get_multiple_fields(key_value=TEST_ACCOUNT_ID, getters={
            'one': FieldGetter(field_path='multiAttributesContainer.{{itemKey}}.(name, value)', query_kwargs=query_kwargs)
        })
        self.assertEqual(retrieved_values['one'].get('name', None), random_name)
        self.assertEqual(retrieved_values['one'].get('value', None), random_value)

        deletion_success: bool = self.users_table.delete_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='multiAttributesContainer.{{itemKey}}',
            query_kwargs=query_kwargs
        )
        self.assertTrue(deletion_success)


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Any, Type
from StructNoSQL import DynamoDBBasicTable, PrimaryIndex, GlobalSecondaryIndex, DynamoDBCachingTable, \
    ExternalDynamoDBApiCachingTable, ExternalDynamoDBApiBasicTable, TableDataModel


TEST_ACCOUNT_ID = "5ae5938d-d4b5-41a7-ad33-40f3c1476211"
TEST_PROJECT_ID = "defcc77c-1d6d-46a4-8cbe-506d12b824b7"
TEST_ACCOUNT_EMAIL = "yay.com"
TEST_ACCOUNT_USERNAME = "Yay"


class PlaygroundDynamoDBBasicTable(DynamoDBBasicTable):
    def __init__(self, data_model: Optional[Any] = None):
        primary_index = PrimaryIndex(hash_key_name='accountId', hash_key_variable_python_type=str)
        globals_secondary_indexes = [
            GlobalSecondaryIndex(hash_key_name='username', hash_key_variable_python_type=str, projection_type='ALL'),
            GlobalSecondaryIndex(hash_key_name='email', hash_key_variable_python_type=str, projection_type='ALL'),
            GlobalSecondaryIndex(hash_key_name='type', hash_key_variable_python_type=str, projection_type='ALL'),
        ]
        super().__init__(
            table_name="structnosql-playground", region_name="eu-west-2", data_model=data_model,
            primary_index=primary_index, global_secondary_indexes=globals_secondary_indexes, auto_create_table=True
        )

class PlaygroundDynamoDBCachingTable(DynamoDBCachingTable):
    def __init__(self, data_model):
        primary_index = PrimaryIndex(hash_key_name="accountId", hash_key_variable_python_type=str)
        globals_secondary_indexes = [
            GlobalSecondaryIndex(hash_key_name="username", hash_key_variable_python_type=str, projection_type='ALL'),
            GlobalSecondaryIndex(hash_key_name="email", hash_key_variable_python_type=str, projection_type='ALL'),
            GlobalSecondaryIndex(hash_key_name='type', hash_key_variable_python_type=str, projection_type='ALL'),
        ]
        super().__init__(
            table_name="structnosql-playground", region_name="eu-west-2", data_model=data_model,
            primary_index=primary_index, global_secondary_indexes=globals_secondary_indexes,
            auto_create_table=True
        )


class InoftVocalEngineBasicTable(ExternalDynamoDBApiBasicTable):
    def __init__(
            self, data_model: Type[TableDataModel],
            engine_account_id: str, engine_project_id: str, engine_api_key: str,
            table_id: str, region_name: str
    ):
        self.engine_account_id = engine_account_id
        self.engine_project_id = engine_project_id
        super().__init__(
            api_http_endpoint=f'http://127.0.0.1:5000/api/v1/{self.engine_account_id}/{self.engine_project_id}/database-client',
            primary_index=PrimaryIndex(hash_key_name='accountProjectTableKeyId', hash_key_variable_python_type=str),
            data_model=data_model, base_payload={'accessToken': engine_api_key, 'tableId': table_id, 'regionName': region_name}
        )

class InoftVocalEngineCachingTable(ExternalDynamoDBApiCachingTable):
    def __init__(
            self,  data_model: Type[TableDataModel],
            engine_account_id: str, engine_project_id: str, engine_api_key: str,
            table_id: str, region_name: str
    ):
        self.engine_account_id = engine_account_id
        self.engine_project_id = engine_project_id
        super().__init__(
            api_http_endpoint=f'http://127.0.0.1:5000/api/v1/{self.engine_account_id}/{self.engine_project_id}/database-client',
            primary_index=PrimaryIndex(hash_key_name='accountProjectTableKeyId', hash_key_variable_python_type=str),
            data_model=data_model, base_payload={'accessToken': engine_api_key, 'tableId': table_id, 'regionName': region_name}
        )


INOFT_VOCAL_ENGINE_PLAYGROUND_TABLE_KWARGS = {
    'engine_account_id': "b1fe5939-032b-462d-92e0-a942cd445096",
    'engine_project_id': "4ede8b70-46f6-4ae2-b09c-05a549194c8e",
    'engine_api_key': "daisybelle",
    'table_id': "structnosql-playground",
    'region_name': "eu-west-2",
}

class PlaygroundInoftVocalEngineBasicTable(InoftVocalEngineBasicTable):
    def __init__(self, data_model: Type[TableDataModel]):
        super().__init__(**INOFT_VOCAL_ENGINE_PLAYGROUND_TABLE_KWARGS, data_model=data_model)

class PlaygroundInoftVocalEngineCachingTable(InoftVocalEngineCachingTable):
    def __init__(self, data_model: Type[TableDataModel]):
        super().__init__(**INOFT_VOCAL_ENGINE_PLAYGROUND_TABLE_KWARGS, data_model=data_model)



import unittest
from sys import getsizeof
from typing import Dict

from StructNoSQL import FieldRemover, BaseField, TableDataModel
from StructNoSQL.clients_middlewares.dynamodb.backend.dynamodb_core import EXPRESSION_MAX_BYTES_SIZE
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TestExceedOperationsLimitSize(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)

    def test_deletion_and_removal_limit_size(self):
        class TableModel(TableDataModel):
            accountId = BaseField(field_type=str, required=True)
            container = BaseField(field_type=bool, required=False)

        table_model = TableModel()
        removers: Dict[str, FieldRemover] = dict()

        index = 0
        while True:
            current_field_path = f"dummy{index}"
            removers[current_field_path] = FieldRemover(field_path=current_field_path)
            table_model.class_add_field(
                field_key=current_field_path,
                field_item=BaseField(field_type=bool, required=False)
            )
            if getsizeof(removers) > (EXPRESSION_MAX_BYTES_SIZE * 1.5):
                break
            index += 1

        users_table = PlaygroundDynamoDBBasicTable(data_model=table_model)
        # We initialize the table after having added all the fields to the class of the TableModel, so that
        # the indexing of the model will correctly be done on all the programmatically added fields.

        deletion_response = users_table.delete_multiple_fields(key_value=TEST_ACCOUNT_ID, removers=removers)
        self.assertTrue(all(deletion_response.values()))

        retrieved_removed_data = users_table.remove_multiple_fields(key_value=TEST_ACCOUNT_ID, removers=removers)
        self.assertIsNotNone(retrieved_removed_data)


if __name__ == '__main__':
    unittest.main()



import unittest

from tests.components.playground_table_clients import PlaygroundInoftVocalEngineBasicTable
from tests.tests_caching_table.table_models import ExternalDynamoDBApiTableModel


class TestsExternalDynamoDBApiCachingTable(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundInoftVocalEngineBasicTable(data_model=ExternalDynamoDBApiTableModel)
        self.users_table.debug = True

        self.SHARED_CASE_KWARGS = {'self': self, 'users_table': self.users_table, 'primary_key_name': 'accountProjectTableKeyId', 'is_caching': False}

    def test_set_get_fields_with_primary_index(self):
        from tests.tests_query_operations.cases_shared import test_set_get_fields_with_primary_index
        test_set_get_fields_with_primary_index(**self.SHARED_CASE_KWARGS)


if __name__ == '__main__':
    unittest.main()



from typing import Optional, List, Dict, Any, Tuple, Callable, Union, Type

from StructNoSQL import PrimaryIndex, BaseField, TableDataModel
from StructNoSQL.models import DatabasePathElement, FieldGetter, FieldSetter, UnsafeFieldSetter, FieldRemover, \
    FieldPathSetter, QueryMetadata
from StructNoSQL.tables.base_table import BaseTable
from StructNoSQL.tables.shared_table_behaviors import _prepare_getters, _model_contain_all_index_keys, \
    unpack_validate_retrieved_field_if_need_to, unpack_validate_multiple_retrieved_fields_if_need_to
from StructNoSQL.utils.data_processing import navigate_into_data_with_field_path_elements
from StructNoSQL.utils.process_render_fields_paths import process_and_make_single_rendered_database_path,\
    process_validate_data_and_make_single_rendered_database_path


class BaseBasicTable(BaseTable):
    def __init__(self, data_model: Type[TableDataModel], primary_index: PrimaryIndex):
        super().__init__(data_model=data_model, primary_index=primary_index)

    def _put_record(self, middleware: Callable[[dict], bool], record_dict_data: dict) -> bool:
        self.model_virtual_map_field.populate(value=record_dict_data)
        validated_data, is_valid = self.model_virtual_map_field.validate_data()
        return middleware(validated_data) if is_valid is True else False

    def _delete_record(self, middleware: Callable[[dict], bool], indexes_keys_selectors: dict) -> bool:
        found_all_indexes: bool = _model_contain_all_index_keys(model=self.model, indexes_keys=indexes_keys_selectors.keys())
        if found_all_indexes is not True:
            return False
        return middleware(indexes_keys_selectors)

    def _remove_record(
            self, middleware: Callable[[dict], Optional[dict]],
            indexes_keys_selectors: dict, data_validation: bool
    ) -> Optional[dict]:
        found_all_indexes: bool = _model_contain_all_index_keys(model=self.model, indexes_keys=indexes_keys_selectors.keys())
        if found_all_indexes is not True:
            return None

        removed_record_data: Optional[dict] = middleware(indexes_keys_selectors)
        if removed_record_data is None:
            return None

        if data_validation is True:
            self.model_virtual_map_field.populate(value=removed_record_data)
            validated_data, is_valid = self.model_virtual_map_field.validate_data()
            return validated_data
        else:
            return removed_record_data

    def _get_field(
            self, middleware: Callable[[Union[List[DatabasePathElement], Dict[str, List[DatabasePathElement]]], bool], Optional[Any]],
            field_path: str, query_kwargs: Optional[dict], data_validation: bool
    ) -> Optional[Any]:
        target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs
        )
        if is_multi_selector is not True:
            target_field_container: Tuple[BaseField, List[DatabasePathElement]]
            field_object, field_path_elements = target_field_container

            retrieved_item_data: Optional[Any] = middleware(field_path_elements, False)
            if data_validation is not True:
                return retrieved_item_data
            else:
                field_object.populate(value=retrieved_item_data)
                validated_data, is_valid = field_object.validate_data()
                return validated_data
        else:
            target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]

            fields_paths_elements: Dict[str, List[DatabasePathElement]] = {key: item[1] for key, item in target_field_container.items()}
            retrieved_items_data: Dict[str, Optional[Any]] = middleware(fields_paths_elements, True)

            output_data: Dict[str, Optional[Any]] = {}
            for item_key, item_container in target_field_container.items():
                field_object, field_path_elements = item_container

                matching_item_data: Optional[Any] = retrieved_items_data.get(item_key, None)
                if data_validation is not True:
                    output_data[item_key] = matching_item_data
                else:
                    field_object.populate(value=matching_item_data)
                    validated_data, is_valid = field_object.validate_data()
                    output_data[item_key] = validated_data

            return output_data

    def inner_query_fields_secondary_index(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Tuple[Optional[List[Any]], QueryMetadata]],
            fields_database_paths: List[List[DatabasePathElement]],
    ) -> Tuple[Optional[dict], QueryMetadata]:
        from StructNoSQL.tables.shared_table_behaviors import _inner_query_fields_secondary_index
        return _inner_query_fields_secondary_index(
            primary_index_name=self.primary_index_name,
            get_primary_key_database_path=self._get_primary_key_database_path,
            middleware=middleware,
            fields_paths_elements=fields_database_paths,
        )

    @staticmethod
    def unpack_validate_getters_record_attributes_if_need_to(
            data_validation: bool, record_attributes: dict,
            single_getters_target_fields_containers: Dict[str, Tuple[BaseField, List[DatabasePathElement]]],
            grouped_getters_target_fields_containers: Dict[str, Dict[str, Tuple[BaseField, List[DatabasePathElement]]]],
    ):
        def item_mutator(item_value: Any, item_field_path_elements: List[DatabasePathElement]) -> Any:
            return item_value

        from StructNoSQL.tables.shared_table_behaviors import _unpack_validate_getters_record_attributes_if_need_to
        return _unpack_validate_getters_record_attributes_if_need_to(
            item_mutator=item_mutator, data_validation=data_validation, record_attributes=record_attributes,
            single_getters_target_fields_containers=single_getters_target_fields_containers,
            grouped_getters_target_fields_containers=grouped_getters_target_fields_containers
        )

    def _query_field(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Tuple[Optional[List[Any]], QueryMetadata]],
            key_value: str, field_path: str, query_kwargs: Optional[dict], index_name: Optional[str], data_validation: bool
    ) -> Tuple[Optional[dict], QueryMetadata]:

        target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs
        )
        if index_name is not None and index_name != self.primary_index_name:
            output_records_values: dict = {}
            if is_multi_selector is not True:
                target_field_container: Tuple[BaseField, List[DatabasePathElement]]
                fields_database_paths: List[List[DatabasePathElement]] = [target_field_container[1]]

                records_attributes, query_metadata = self.inner_query_fields_secondary_index(
                    middleware=middleware, fields_database_paths=fields_database_paths,
                )
                for record_key, record_item_attributes in records_attributes.items():
                    output_records_values[record_key] = unpack_validate_retrieved_field_if_need_to(
                        record_attributes=record_item_attributes, 
                        target_field_container=target_field_container,
                        data_validation=data_validation
                    )
            else:
                target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]
                fields_database_paths: List[List[DatabasePathElement]] = [item[1] for item in target_field_container.values()]

                records_attributes, query_metadata = self.inner_query_fields_secondary_index(
                    middleware=middleware, fields_database_paths=fields_database_paths,
                )
                for record_key, record_item_attributes in records_attributes.items():
                    output_records_values[record_key] = unpack_validate_multiple_retrieved_fields_if_need_to(
                        record_attributes=record_item_attributes,
                        target_fields_containers=target_field_container,
                        data_validation=data_validation
                    )
            return output_records_values, query_metadata
        else:
            # If requested index is primary index
            if is_multi_selector is not True:
                target_field_container: Tuple[BaseField, List[DatabasePathElement]]
                field_object, field_path_elements = target_field_container

                retrieved_records_items_data, query_metadata = middleware([field_path_elements])
                if retrieved_records_items_data is not None and len(retrieved_records_items_data) > 0:
                    # Since we query the primary_index, we know for a fact that we will never be returned more than
                    # one record item. Hence why we do not have a loop that iterate over the records_items_data,
                    # and that we return a dict with the only one record item being the requested key_value.
                    return {key_value: unpack_validate_retrieved_field_if_need_to(
                        data_validation=data_validation,
                        record_attributes=retrieved_records_items_data[0],
                        target_field_container=target_field_container
                    )}, query_metadata
                return None, query_metadata
            else:
                target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]

                fields_paths_elements: List[List[DatabasePathElement]] = [item[1] for item in target_field_container.values()]
                retrieved_records_items_data, query_metadata = middleware(fields_paths_elements)
                if retrieved_records_items_data is not None and len(retrieved_records_items_data) > 0:
                    # Since we query the primary_index, we know for a fact that we will never be returned more than
                    # one record item. Hence why we do not have a loop that iterate over the records_items_data,
                    # and that we return a dict with the only one record item being the requested key_value.
                    return {key_value: unpack_validate_multiple_retrieved_fields_if_need_to(
                        data_validation=data_validation,
                        record_attributes=retrieved_records_items_data[0],
                        target_fields_containers=target_field_container
                    )}, query_metadata
                return None, query_metadata

    def _query_multiple_fields(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Tuple[Optional[List[Any]], QueryMetadata]],
            key_value: str, getters: Dict[str, FieldGetter], index_name: Optional[str], data_validation: bool
    ) -> Tuple[Optional[dict], QueryMetadata]:
        getters_database_paths, single_getters_target_fields_containers, grouped_getters_target_fields_containers = (
            _prepare_getters(fields_switch=self.fields_switch, getters=getters)
        )

        if index_name is None or index_name == self.primary_index_name:
            retrieved_records_items_data, query_metadata = middleware(getters_database_paths)
            if retrieved_records_items_data is not None and len(retrieved_records_items_data) > 0:
                # Since we query the primary_index, we know for a fact that we will never be returned more than
                # one record item. Hence why we do not have a loop that iterate over the records_items_data,
                # and that we return a dict with the only one record item being the requested key_value.
                record_item_value: Dict[str, Any] = self.unpack_validate_getters_record_attributes_if_need_to(
                    data_validation=data_validation, record_attributes=retrieved_records_items_data[0],
                    single_getters_target_fields_containers=single_getters_target_fields_containers,
                    grouped_getters_target_fields_containers=grouped_getters_target_fields_containers
                )
                return {key_value: record_item_value}, query_metadata
            return None, query_metadata
        else:
            records_attributes, query_metadata = self.inner_query_fields_secondary_index(
                middleware=middleware, fields_database_paths=getters_database_paths,
            )
            output_records_values: dict = {
                record_key: self.unpack_validate_getters_record_attributes_if_need_to(
                    data_validation=data_validation, record_attributes=record_item_attributes,
                    single_getters_target_fields_containers=single_getters_target_fields_containers,
                    grouped_getters_target_fields_containers=grouped_getters_target_fields_containers
                ) for record_key, record_item_attributes in records_attributes.items()
            }
            return output_records_values, query_metadata

    def _get_multiple_fields(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Any],
            getters: Dict[str, FieldGetter], data_validation: bool
    ) -> Dict[str, Optional[Any]]:

        getters_database_paths, single_getters_target_fields_containers, grouped_getters_target_fields_containers = (
            _prepare_getters(fields_switch=self.fields_switch, getters=getters)
        )
        record_attributes: Optional[dict] = middleware(getters_database_paths)
        if record_attributes is None:
            return {getter_key: None for getter_key in getters.keys()}

        return self.unpack_validate_getters_record_attributes_if_need_to(
            data_validation=data_validation, record_attributes=record_attributes,
            single_getters_target_fields_containers=single_getters_target_fields_containers,
            grouped_getters_target_fields_containers=grouped_getters_target_fields_containers,
        )

    def _update_field(
            self, middleware: Callable[[List[DatabasePathElement], Any], bool],
            field_path: str, value_to_set: Any, query_kwargs: Optional[dict] = None
    ) -> bool:
        field_object, field_path_elements, validated_data, is_valid = process_validate_data_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs, data_to_validate=value_to_set
        )
        return middleware(field_path_elements, validated_data) if is_valid is True else False

    def _update_field_return_old(
            self, middleware: Callable[[List[DatabasePathElement], Any], Tuple[bool, Any]],
            field_path: str, value_to_set: Any, query_kwargs: Optional[dict], data_validation: bool
    ) -> Tuple[bool, Optional[Any]]:
        field_object, field_path_elements, validated_update_data, update_data_is_valid = process_validate_data_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs, data_to_validate=value_to_set
        )
        if update_data_is_valid is not True:
            return False, None

        update_success, response_attributes = middleware(field_path_elements, validated_update_data)

        from StructNoSQL.utils.data_processing import navigate_into_data_with_field_path_elements
        old_item_data: Optional[Any] = navigate_into_data_with_field_path_elements(
            data=response_attributes, field_path_elements=field_path_elements,
            num_keys_to_navigation_into=len(field_path_elements)
        ) if response_attributes is not None else None

        if data_validation is True:
            field_object.populate(value=old_item_data)
            validated_removed_data, removed_data_is_valid = field_object.validate_data()
        else:
            validated_removed_data = old_item_data

        return update_success, validated_removed_data

    def _update_multiple_fields(
            self, middleware: Callable[[List[FieldPathSetter]], bool],
            setters: List[FieldSetter or UnsafeFieldSetter]
    ) -> bool:
        dynamodb_setters: List[FieldPathSetter] = []
        for current_setter in setters:
            if isinstance(current_setter, FieldSetter):
                field_object, field_path_elements, validated_data, is_valid = process_validate_data_and_make_single_rendered_database_path(
                    field_path=current_setter.field_path, fields_switch=self.fields_switch,
                    query_kwargs=current_setter.query_kwargs, data_to_validate=current_setter.value_to_set
                )
                if is_valid is True:
                    dynamodb_setters.append(FieldPathSetter(
                        field_path_elements=field_path_elements, value_to_set=validated_data
                    ))
            elif isinstance(current_setter, UnsafeFieldSetter):
                raise Exception(f"UnsafeFieldSetter not supported in basic_table")

                """safe_field_path_object, is_multi_selector = process_and_get_field_path_object_from_field_path(
                    field_path_key=current_setter.safe_base_field_path, fields_switch=self.fields_switch
                )
                # todo: add support for multiple fields path
                if current_setter.unsafe_path_continuation is None:
                    field_path_elements = safe_field_path_object.database_path
                else:
                    field_path_elements = safe_field_path_object.database_path + current_setter.unsafe_path_continuation

                processed_value_to_set: Any = float_to_decimal_serializer(current_setter.value_to_set)
                # Since the data is not validated, we pass it to the float_to_decimal_serializer
                # function (which normally should be called by the data validation function)

                rendered_field_path_elements = make_rendered_database_path(
                    database_path_elements=field_path_elements,
                    query_kwargs=current_setter.query_kwargs
                )
                dynamodb_setters.append(FieldPathSetter(
                    field_path_elements=rendered_field_path_elements,
                    value_to_set=processed_value_to_set
                ))"""
        update_success: bool = middleware(dynamodb_setters)
        return update_success

    def _update_multiple_fields_return_old(
            self, middleware: Callable[[Dict[str, FieldPathSetter]], Tuple[bool, Dict[str, Optional[Any]]]],
            setters: Dict[str, FieldSetter], data_validation: bool
    ) -> Tuple[bool, Dict[str, Optional[Any]]]:

        setters_containers: Dict[str, Tuple[BaseField, List[DatabasePathElement]]] = {}
        dynamodb_setters: Dict[str, FieldPathSetter] = {}
        for setter_key, setter_item in setters.items():
            field_object, field_path_elements, validated_data, is_valid = process_validate_data_and_make_single_rendered_database_path(
                field_path=setter_item.field_path, fields_switch=self.fields_switch,
                query_kwargs=setter_item.query_kwargs, data_to_validate=setter_item.value_to_set
            )
            if is_valid is True:
                setters_containers[setter_key] = (field_object, field_path_elements)
                dynamodb_setters[setter_key] = FieldPathSetter(
                    field_path_elements=field_path_elements, value_to_set=validated_data
                )
        update_success, setters_response_attributes = middleware(dynamodb_setters)

        output_data: Dict[str, Optional[Any]] = {}
        for item_key, item_container in setters_containers.items():
            item_field_object, item_field_path_elements = item_container

            item_data: Optional[Any] = navigate_into_data_with_field_path_elements(
                data=setters_response_attributes, field_path_elements=item_field_path_elements,
                num_keys_to_navigation_into=len(item_field_path_elements)
            )
            if data_validation is True:
                item_field_object.populate(value=item_data)
                validated_data, valid = item_field_object.validate_data()
                output_data[item_key] = validated_data
            else:
                output_data[item_key] = item_data

        return update_success, output_data

    def _remove_field(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Optional[dict]],
            field_path: str, query_kwargs: Optional[dict], data_validation: bool
    ) -> Optional[Any]:
        target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs
        )

        if is_multi_selector is not True:
            target_field_container: Tuple[BaseField, List[DatabasePathElement]]
            field_path_object, field_path_elements = target_field_container

            removed_item_attributes: Optional[dict] = middleware([field_path_elements])
            if removed_item_attributes is None:
                return None

            item_removed_data: Optional[Any] = navigate_into_data_with_field_path_elements(
                data=removed_item_attributes, field_path_elements=field_path_elements,
                num_keys_to_navigation_into=len(field_path_elements)
            )
            if data_validation is True:
                field_path_object.populate(value=item_removed_data)
                validated_data, is_valid = field_path_object.validate_data()
                return validated_data
            else:
                return item_removed_data
        else:
            target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]

            fields_paths_elements: List[List[DatabasePathElement]] = [item[1] for item in target_field_container.values()]
            removed_items_attributes: Optional[dict] = middleware(fields_paths_elements)
            # The attributes of all the removed items are packed inside the same dictionary,
            # because all the remove's are expected to be done by a single database operation.
            if removed_items_attributes is None:
                return None

            removed_items_values: Dict[str, Optional[Any]] = {}
            for item_key, item_container in target_field_container.items():
                item_field_object, item_field_path_elements = item_container

                item_removed_data: Optional[Any] = navigate_into_data_with_field_path_elements(
                    data=removed_items_attributes, field_path_elements=item_field_path_elements,
                    num_keys_to_navigation_into=len(item_field_path_elements)
                )
                if data_validation is True:
                    item_field_object.populate(value=item_removed_data)
                    validated_data, is_valid = item_field_object.validate_data()
                    removed_items_values[item_key] = validated_data
                else:
                    removed_items_values[item_key] = item_removed_data

            return removed_items_values

    def _delete_field(
            self, middleware: Callable[[List[List[DatabasePathElement]]], bool],
            field_path: str, query_kwargs: Optional[dict] = None
    ) -> bool:
        target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs
        )
        if is_multi_selector is not None:
            target_field_container: Tuple[BaseField, List[DatabasePathElement]]
            return middleware([target_field_container[1]])
        else:
            target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]
            targets_paths_elements: List[List[DatabasePathElement]] = [item[1] for item in target_field_container.values()]
            return middleware(targets_paths_elements)

    def _grouped_remove_multiple_fields(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Any],
            removers: Dict[str, FieldRemover], data_validation: bool
    ) -> Optional[Dict[str, Any]]:
        if not len(removers) > 0:
            # If no remover has been specified, we do not run the database
            # operation, and since no value has been removed, we return None.
            return None

        removers_field_paths_elements: Dict[str, Tuple[BaseField, List[DatabasePathElement]]] = {}
        grouped_removers_field_paths_elements: Dict[str, Dict[str, Tuple[BaseField, List[DatabasePathElement]]]] = {}

        removers_database_paths: List[List[DatabasePathElement]] = []
        for remover_key, remover_item in removers.items():
            target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
                field_path=remover_item.field_path, fields_switch=self.fields_switch,
                query_kwargs=remover_item.query_kwargs
            )
            if is_multi_selector is not True:
                target_field_container: Tuple[BaseField, List[DatabasePathElement]]
                field_object, field_path_elements = target_field_container

                removers_field_paths_elements[remover_key] = target_field_container
                removers_database_paths.append(field_path_elements)
            else:
                target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]
                fields_paths_elements: List[List[DatabasePathElement]] = [item[1] for item in target_field_container.values()]

                grouped_removers_field_paths_elements[remover_key] = target_field_container
                removers_database_paths.extend(fields_paths_elements)

        record_attributes: Optional[dict] = middleware(removers_database_paths)
        if record_attributes is None:
            return None

        return self.unpack_validate_getters_record_attributes_if_need_to(
            data_validation=data_validation, record_attributes=record_attributes,
            single_getters_target_fields_containers=removers_field_paths_elements,
            grouped_getters_target_fields_containers=grouped_removers_field_paths_elements
        )

    def _grouped_delete_multiple_fields(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Any], removers: List[FieldRemover],
    ) -> bool:
        if not len(removers) > 0:
            # If no remover has been specified, we do not run the database operation, yet we still
            # return True, since technically, what needed to be performed (nothing) was performed.
            return True

        removers_database_paths: List[List[DatabasePathElement]] = []
        for current_remover in removers:
            target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
                field_path=current_remover.field_path, fields_switch=self.fields_switch,
                query_kwargs=current_remover.query_kwargs
            )
            if is_multi_selector is not True:
                target_field_container: Tuple[BaseField, List[DatabasePathElement]]
                removers_database_paths.append(target_field_container[1])
            else:
                target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]
                fields_paths_elements: List[List[DatabasePathElement]] = [item[1] for item in target_field_container.values()]
                removers_database_paths.extend(fields_paths_elements)

        response: Optional[Any] = middleware(removers_database_paths)
        return response is not None



from StructNoSQL import TableDataModel, BaseField, MapModel


class BaseFirstTableModel(TableDataModel):
    simpleField = BaseField(field_type=str, required=False)
    class ContainerModel(MapModel):
        nestedFieldOne = BaseField(field_type=str, required=False)
        nestedFieldTwo = BaseField(field_type=str, required=False)
    container = BaseField(field_type=ContainerModel, required=False)

class BaseSecondTableModel(TableDataModel):
    simpleField = BaseField(field_type=int, required=False)
    class ContainerModel(MapModel):
        nestedFieldOne = BaseField(field_type=int, required=False)
        nestedFieldTwo = BaseField(field_type=int, required=False)
    container = BaseField(field_type=ContainerModel, required=False)


class DynamoDBFirstTableModel(BaseFirstTableModel):
    accountId = BaseField(field_type=str, required=True)
class DynamoDBSecondTableModel(BaseSecondTableModel):
    accountId = BaseField(field_type=str, required=True)

class ExternalDynamoDBApiFirstTableModel(BaseFirstTableModel):
    accountProjectTableKeyId = BaseField(field_type=str, required=True)
class ExternalDynamoDBApiSecondTableModel(BaseSecondTableModel):
    accountProjectTableKeyId = BaseField(field_type=str, required=True)



from typing import Optional, List, Dict, Any, Tuple, Union, Generator, Type

from StructNoSQL import TableDataModel
from StructNoSQL.clients_middlewares.dynamodb.backend.dynamodb_core import DynamoDbCoreAdapter, PrimaryIndex, GlobalSecondaryIndex
from StructNoSQL.clients_middlewares.dynamodb.backend.dynamodb_utils import DynamoDBUtils
from StructNoSQL.clients_middlewares.dynamodb.backend.models import Response
from StructNoSQL.models import DatabasePathElement, FieldGetter, FieldSetter, UnsafeFieldSetter, FieldRemover, \
    FieldPathSetter, QueryMetadata
from StructNoSQL.practical_logger import message_with_vars
from StructNoSQL.tables.base_basic_table import BaseBasicTable
from StructNoSQL.clients_middlewares.dynamodb.dynamodb_low_level_table_operations import DynamoDBLowLevelTableOperations
from StructNoSQL.tables.shared_table_behaviors import _prepare_getters


class DynamoDBBasicTable(BaseBasicTable, DynamoDBLowLevelTableOperations):
    def __init__(
            self, table_name: str, region_name: str,
            data_model: Type[TableDataModel], primary_index: PrimaryIndex,
            billing_mode: str = DynamoDbCoreAdapter.PAY_PER_REQUEST,
            global_secondary_indexes: List[GlobalSecondaryIndex] = None,
            auto_create_table: bool = True
    ):
        super().__init__(data_model=data_model, primary_index=primary_index)
        super().__setup_connectors__(
            table_name=table_name, region_name=region_name,
            primary_index=primary_index, global_secondary_indexes=global_secondary_indexes,
            billing_mode=billing_mode, auto_create_table=auto_create_table
        )

    def put_record(self, record_dict_data: dict) -> bool:
        def middleware(validated_record_item: dict) -> bool:
            return self.dynamodb_client.put_record(item_dict=validated_record_item)
        return self._put_record(middleware=middleware, record_dict_data=record_dict_data)

    def delete_record(self, indexes_keys_selectors: dict) -> bool:
        def middleware(indexes_keys: dict) -> bool:
            return self.dynamodb_client.delete_record(indexes_keys_selectors=indexes_keys)
        return self._delete_record(middleware=middleware, indexes_keys_selectors=indexes_keys_selectors)

    def remove_record(self, indexes_keys_selectors: dict, data_validation: bool = True) -> Optional[dict]:
        def middleware(indexes_keys: dict) -> Optional[dict]:
            return self.dynamodb_client.remove_record(indexes_keys_selectors=indexes_keys)
        return self._remove_record(middleware=middleware, indexes_keys_selectors=indexes_keys_selectors, data_validation=data_validation)

    def get_field(self, key_value: str, field_path: str, query_kwargs: Optional[dict] = None, index_name: Optional[str] = None, data_validation: bool = True) -> Any:
        def middleware(field_path_elements: Union[List[DatabasePathElement], Dict[str, List[DatabasePathElement]]], is_multi_selector: bool):
            if is_multi_selector is not True:
                field_path_elements: List[DatabasePathElement]
                response_data: Optional[Any] = self.dynamodb_client.get_value_in_path_target(
                    index_name=index_name or self.primary_index_name,
                    key_value=key_value, field_path_elements=field_path_elements
                )
                return response_data
            else:
                field_path_elements: Dict[str, List[DatabasePathElement]]
                response_data: Optional[Dict[str, Any]] = self.dynamodb_client.get_values_in_multiple_path_target(
                    index_name=index_name or self.primary_index_name,
                    key_value=key_value, fields_path_elements=field_path_elements
                )
                if response_data is not None:
                    return {key: response_data.get(key, None) for key in field_path_elements.keys()}
                else:
                    # When using multi-selectors, the middleware is always expected to return a dict with all the
                    # 'getters' keys present, hence why the dict with None values if the response_data is None.
                    return {key: None for key in field_path_elements.keys()}
        return self._get_field(middleware=middleware, field_path=field_path, query_kwargs=query_kwargs, data_validation=data_validation)

    def get_multiple_fields(self, key_value: str, getters: Dict[str, FieldGetter], index_name: Optional[str] = None, data_validation: bool = True) -> Optional[dict]:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]):
            return self.dynamodb_client.get_or_query_single_item(
                index_name=index_name or self.primary_index_name,
                key_value=key_value, fields_path_elements=fields_path_elements,
            )
        return self._get_multiple_fields(middleware=middleware, getters=getters, data_validation=data_validation)

    def query_field(
            self, key_value: str, field_path: str, query_kwargs: Optional[dict] = None, index_name: Optional[str] = None,
            filter_expression: Optional[Any] = None, pagination_records_limit: Optional[int] = None, exclusive_start_key: Optional[Any] = None,
            data_validation: bool = True, **additional_kwargs
    ) -> Tuple[Optional[dict], QueryMetadata]:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]) -> Tuple[Optional[List[dict]], QueryMetadata]:
            return self.dynamodb_client.query_items_by_key(
                index_name=index_name or self.primary_index_name,
                key_value=key_value, fields_path_elements=fields_path_elements,
                pagination_records_limit=pagination_records_limit, filter_expression=filter_expression,
                exclusive_start_key=exclusive_start_key,
                **additional_kwargs
            )
        return self._query_field(
            middleware=middleware, key_value=key_value, field_path=field_path,
            query_kwargs=query_kwargs, index_name=index_name, data_validation=data_validation
        )

    def paginated_query_field(
            self, key_value: str, field_path: str, query_kwargs: Optional[dict] = None, index_name: Optional[str] = None,
            filter_expression: Optional[Any] = None, pagination_records_limit: Optional[int] = None, exclusive_start_key: Optional[Any] = None,
            data_validation: bool = True, **additional_kwargs
    ) -> Generator[Tuple[Optional[dict], QueryMetadata], None, None]:
        current_exclusive_start_key: Optional[Any] = exclusive_start_key
        while True:
            records_data, query_metadata = self.query_field(
                index_name=index_name, key_value=key_value, field_path=field_path, query_kwargs=query_kwargs,
                pagination_records_limit=pagination_records_limit, filter_expression=filter_expression,
                exclusive_start_key=current_exclusive_start_key, data_validation=data_validation, **additional_kwargs
            )
            yield records_data, query_metadata
            if query_metadata.last_evaluated_key is None:
                break
            current_exclusive_start_key = query_metadata.last_evaluated_key

    def query_multiple_fields(
            self, key_value: str, getters: Dict[str, FieldGetter], index_name: Optional[str] = None,
            filter_expression: Optional[Any] = None, pagination_records_limit: Optional[int] = None, exclusive_start_key: Optional[Any] = None,
            data_validation: bool = True, **additional_kwargs
    ) -> Tuple[Optional[Dict[str, dict]], QueryMetadata]:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]) -> Tuple[List[dict], QueryMetadata]:
            return self.dynamodb_client.query_items_by_key(
                index_name=index_name or self.primary_index_name, key_value=key_value,
                fields_path_elements=fields_path_elements,
                pagination_records_limit=pagination_records_limit, filter_expression=filter_expression,
                exclusive_start_key=exclusive_start_key,
                **additional_kwargs
            )
        return self._query_multiple_fields(
            middleware=middleware, key_value=key_value, getters=getters,
            index_name=index_name, data_validation=data_validation
        )

    def paginated_query_multiple_fields(
            self, key_value: str, getters: Dict[str, FieldGetter], index_name: Optional[str] = None,
            filter_expression: Optional[Any] = None, pagination_records_limit: Optional[int] = None, exclusive_start_key: Optional[Any] = None,
            data_validation: bool = True, **additional_kwargs
    ) -> Generator[Tuple[Optional[Dict[str, dict]], QueryMetadata], None, None]:
        current_exclusive_start_key: Optional[Any] = exclusive_start_key
        while True:
            records_data, query_metadata = self.query_multiple_fields(
                key_value=key_value, getters=getters, index_name=index_name,
                pagination_records_limit=pagination_records_limit,
                filter_expression=filter_expression,
                exclusive_start_key=current_exclusive_start_key,
                data_validation=data_validation, **additional_kwargs
            )
            yield records_data, query_metadata
            if query_metadata.last_evaluated_key is None:
                break
            current_exclusive_start_key = query_metadata.last_evaluated_key

    def lightweight_query_multiple_fields(
            self, key_value: str, getters: Dict[str, FieldGetter], index_name: Optional[str] = None,
            pagination_records_limit: Optional[int] = None, filter_expression: Optional[Any] = None, **additional_kwargs
    ) -> Optional[List[dict]]:

        getters_database_paths, single_getters_database_paths_elements, grouped_getters_database_paths_elements = (
            _prepare_getters(fields_switch=self.fields_switch, getters=getters)
        )
        response = self.dynamodb_client.query_response_by_key(
            index_name=index_name or self.primary_index_name, key_value=key_value,
            fields_path_elements=getters_database_paths,
            pagination_records_limit=pagination_records_limit, filter_expression=filter_expression,
            **additional_kwargs
        )
        if response is None:
            return None

        output: List[Dict[str, Any]] = [
            self._unpack_getters_response_item(
                response_item=record_item_data,
                single_getters_database_paths_elements=single_getters_database_paths_elements,
                grouped_getters_database_paths_elements=grouped_getters_database_paths_elements
            )
            # todo: add data validation
            for record_item_data in response.items
            if isinstance(record_item_data, dict)
        ]
        return output

    def update_field(self, key_value: str, field_path: str, value_to_set: Any, query_kwargs: Optional[dict] = None) -> bool:
        def middleware(field_path_elements: List[DatabasePathElement], validated_data: Any):
            response: Optional[Response] = self.dynamodb_client.set_update_data_element_to_map_with_default_initialization(
                index_name=self.primary_index_name,
                key_value=key_value, value=validated_data,
                field_path_elements=field_path_elements
            )
            return response is not None
        return self._update_field(middleware=middleware, field_path=field_path, value_to_set=value_to_set, query_kwargs=query_kwargs)

    def update_field_return_old(
            self, key_value: str, field_path: str, value_to_set: Any,
            query_kwargs: Optional[dict] = None, data_validation: bool = True
    ) -> Tuple[bool, Optional[Any]]:
        def middleware(field_path_elements: List[DatabasePathElement], validated_data: Any) -> Tuple[bool, Optional[dict]]:
            response: Optional[Response] = self.dynamodb_client.set_update_data_element_to_map_with_default_initialization(
                index_name=self.primary_index_name,
                key_value=key_value, value=validated_data,
                field_path_elements=field_path_elements,
                return_old_value=True
            )
            if response is None:
                return False, None

            python_response_attributes: Optional[dict] = (
                DynamoDBUtils.dynamodb_to_python_higher_level(response.attributes)
                if response.attributes is not None else None
            )
            return True, python_response_attributes
        return self._update_field_return_old(
            middleware=middleware, field_path=field_path, value_to_set=value_to_set,
            query_kwargs=query_kwargs, data_validation=data_validation
        )

    def update_multiple_fields(self, key_value: str, setters: List[FieldSetter or UnsafeFieldSetter]) -> bool:
        def middleware(dynamodb_setters: List[FieldPathSetter]) -> bool:
            response: Optional[Response] = self.dynamodb_client.set_update_multiple_data_elements_to_map(
                index_name=self.primary_index_name, key_value=key_value,
                setters=dynamodb_setters, return_old_values=False
            )
            return response is not None
        return self._update_multiple_fields(middleware=middleware, setters=setters)

    def update_multiple_fields_return_old(self, key_value: str, setters: Dict[str, FieldSetter], data_validation: bool = True) -> Tuple[bool, Dict[str, Optional[Any]]]:
        def middleware(dynamodb_setters: Dict[str, FieldPathSetter]) -> Tuple[bool, Optional[dict]]:
            response: Optional[Response] = self.dynamodb_client.set_update_multiple_data_elements_to_map(
                index_name=self.primary_index_name, key_value=key_value,
                setters=list(dynamodb_setters.values()), return_old_values=True
            )
            if response is None:
                return False, None

            python_response_attributes: Optional[dict] = (
                DynamoDBUtils.dynamodb_to_python_higher_level(response.attributes)
                if response.attributes is not None else None
            )
            return True, python_response_attributes
        return self._update_multiple_fields_return_old(middleware=middleware, setters=setters, data_validation=data_validation)

    def remove_field(self, key_value: str, field_path: str, query_kwargs: Optional[dict] = None, data_validation: bool = True) -> Optional[Any]:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]) -> Optional[Dict[str, Any]]:
            return self.dynamodb_client.remove_data_elements_from_map(
                index_name=self.primary_index_name,
                key_value=key_value, targets_path_elements=fields_path_elements,
                retrieve_removed_elements=True
            )
        return self._remove_field(middleware=middleware, field_path=field_path, query_kwargs=query_kwargs, data_validation=data_validation)

    def remove_multiple_fields(self, key_value: str, removers: Dict[str, FieldRemover], data_validation: bool = True) -> Dict[str, Optional[Any]]:
        def task_executor(remover_item: FieldRemover):
            return self.remove_field(
                key_value=key_value,
                field_path=remover_item.field_path,
                query_kwargs=remover_item.query_kwargs,
                data_validation=data_validation
            )
        return self._async_field_removers_executor(task_executor=task_executor, removers=removers)

    def delete_field(self, key_value: str, field_path: str, query_kwargs: Optional[dict] = None) -> bool:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]) -> bool:
            removed_items = self.dynamodb_client.remove_data_elements_from_map(
                index_name=self.primary_index_name,
                key_value=key_value, targets_path_elements=fields_path_elements,
                retrieve_removed_elements=False
            )
            return removed_items is not None
        return self._delete_field(middleware=middleware, field_path=field_path, query_kwargs=query_kwargs)

    def delete_multiple_fields(self, key_value: str, removers: Dict[str, FieldRemover]) -> Dict[str, bool]:
        def task_executor(remover_item: FieldRemover):
            return self.delete_field(
                key_value=key_value,
                field_path=remover_item.field_path,
                query_kwargs=remover_item.query_kwargs
            )
        return self._async_field_removers_executor(task_executor=task_executor, removers=removers)

    def grouped_remove_multiple_fields(self, key_value: str, removers: Dict[str, FieldRemover], data_validation: bool = True) -> Optional[Dict[str, Any]]:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]):
            return self.dynamodb_client.remove_data_elements_from_map(
                index_name=self.primary_index_name,
                key_value=key_value, targets_path_elements=fields_path_elements,
                retrieve_removed_elements=True
            )
        return self._grouped_remove_multiple_fields(middleware=middleware, removers=removers, data_validation=data_validation)

    def grouped_delete_multiple_fields(self, key_value: str, removers: List[FieldRemover]) -> bool:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]):
            return self.dynamodb_client.remove_data_elements_from_map(
                index_name=self.primary_index_name,
                key_value=key_value, targets_path_elements=fields_path_elements,
                retrieve_removed_elements=False
            )
        return self._grouped_delete_multiple_fields(middleware=middleware, removers=removers)



import unittest
from typing import Dict

from StructNoSQL import TableDataModel, BaseField, MapModel, FieldSetter, FieldRemover, FieldGetter
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    class ItemModel(MapModel):
        itemName = BaseField(field_type=str, required=False)
    containerForItemsWithSharedParents = BaseField(field_type=Dict[str, ItemModel], key_name='itemId', required=False)


class InitializeFieldsWithSharedParents(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_initialize_two_items_with_share_parents(self):
        base_field_path = 'containerForItemsWithSharedParents.{{itemId}}'
        item_name_field_path = f'{base_field_path}.itemName'

        potential_old_data_deletion_success = self.users_table.delete_multiple_fields(key_value=TEST_ACCOUNT_ID, removers={
            'one': FieldRemover(field_path=base_field_path, query_kwargs={'itemId': 'itemOne'}),
            'two': FieldRemover(field_path=base_field_path, query_kwargs={'itemId': 'itemTwo'})
        })
        # We do not check the success of the deletion operation, because at the time of writing that, if a delete operation try to delete
        # a field path that does not exist, the operation will crash and return a success of False, where what we want is just to make sure
        # the data is fully clean so we can make sure we initialize the items from scratch. We do not really care about removing the data.

        update_success = self.users_table.update_multiple_fields(
            key_value=TEST_ACCOUNT_ID, setters=[
                FieldSetter(field_path=item_name_field_path, query_kwargs={'itemId': 'itemOne'}, value_to_set="NameItemOne"),
                FieldSetter(field_path=item_name_field_path, query_kwargs={'itemId': 'itemTwo'}, value_to_set="NameItemTwo"),
            ]
        )
        self.assertTrue(update_success)

        retrieved_items_data = self.users_table.get_multiple_fields(key_value=TEST_ACCOUNT_ID, getters={
            'one': FieldGetter(field_path=base_field_path, query_kwargs={'itemId': 'itemOne'}),
            'two': FieldGetter(field_path=base_field_path, query_kwargs={'itemId': 'itemTwo'}),
        })
        self.assertIsNotNone(retrieved_items_data)
        self.assertEqual(retrieved_items_data.get('one', None), {'itemName': "NameItemOne"})
        self.assertEqual(retrieved_items_data.get('two', None), {'itemName': "NameItemTwo"})


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Any
from pydantic import BaseModel
from StructNoSQL import NoneType


class SerializedType(str):
    _keys_to_types = {'str': str, 'int': int, 'float': float, 'dict': dict, 'list': list, 'NoneType': NoneType}

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value) -> type:
        if not isinstance(value, str):
            raise TypeError('string required')
        matching_type: Optional[type] = SerializedType._keys_to_types.get(value, None)
        if matching_type is None:
            raise TypeError('invalid type')
        return matching_type

class FieldPathElementItemModel(BaseModel):
    elementKey: str
    defaultType: SerializedType or type
    customDefaultValue: Optional[Any] = None
